/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.9 $";
#endif

/*
 * Device Add routines
 */

#include "defs.h"

/*
 * Dump rough path tree.  Used for debug
 */
static int DevDumpPath(DevInfo, Depth, Width)
     DevInfo_t		       *DevInfo;
     int			Depth;
     int			Width;
{
    int				nl = 0;

    if (!DevInfo)
	return 0;

    if (DevInfo->Name) {
	if (Depth > 0)
	    printf("->");
	printf("%s[%d,%d]", PRTS(DevInfo->Name), Depth, Width);
    }

    if (DevInfo->Slaves) {
	DevDumpPath(DevInfo->Slaves, Depth+1, Width);
    } else {
	nl = 1;
    }

    if (DevInfo->Next) {
	DevDumpPath(DevInfo->Next, Depth, Width+1);
    } else {
	nl = 1;
    }

    if (nl)
	printf("\n");

    return 0;
}

/*
 * Check to see if DevInfo exists in TreePtr
 * Returns 1 if device exists.
 * Returns 0 if devices does not exist.
 * Returns -1 on error.
 */
static int CheckExists(DevInfo, TreePtr)
     DevInfo_t		       *DevInfo;
     DevInfo_t		      **TreePtr;
{
    static DevFind_t		Find;
    static DevInfo_t	       *Found;

    if (!DevInfo || !TreePtr) {
	SImsg(SIM_DBG, "DevAdd: CheckExists: Bad args passed.");
	return -1;
    }

    if (!*TreePtr) {
	SImsg(SIM_DBG, "DevAdd: CheckExists: TreePtr is empty!");
	/*
	 * This isn't a real error since this could be the root TreePtr node
	 */
	return 0;
    }

#if	DEBUG_DEVADD
    SImsg(SIM_DBG, "DevAdd: CheckExists: Check for <%s>", DevInfo->Name);
#endif

    /*
     * Make sure device hasn't already been added
     */
    DevFindInit(&Find);
    Find.Tree = *TreePtr;
    Find.NodeName = DevInfo->Name;
    Find.NodeID = DevInfo->NodeID;
#if	defined(NO_DUP_DEV_BY_TYPE)
    Find.Flags |= DFF_NOTYPE;
#endif
#if	!defined(DONT_FIND_ON_SERIAL)
    if (!McsiOptSet(MCSIOPT_NOSERIAL))
	Find.Serial = DevInfo->Serial;
#endif	/* DONT_FIND_ON_SERIAL */

    if (Found = DevFind(&Find)) {
	SImsg(SIM_DBG, 
    "DevAdd: <%s> device with the same `%s' already exists; Master = <%s>",
	      DevInfo->Name, Find.Reason,
	      (DevInfo->Master && DevInfo->Master->Name) 
	      ? DevInfo->Master->Name : "?");

	/*
	 * Add check of Capacity.  If the Capacity differs, then this can't
	 * be the same device.  Some RAID arrays will give the same Serial #
	 * for each LUN, but the LUN sizes may be different.  This will catch
	 * that case.  However, if each LUN has the exact same size
	 * (Capacity), then we'll see that as a duplicate and ignore the
	 * 2nd or later devices.
	 */
	if (DevInfo->Capacity && EQ(DevInfo->Capacity, Found->Capacity)) {
	    SImsg(SIM_DBG, "DevAdd: <%s> Capacity (%s vs %s) differs so we won't ignore this device",
		  DevInfo->Name, DevInfo->Capacity, Found->Capacity);
	    return 0;
	} else {
	    SImsg(SIM_DBG, "DevAdd: <%s> Duplicate device will be ignored",
		  DevInfo->Name);
	}

	/*
	 * Assume that the new device has more/better info than the
	 * existing device.
	 */
	if (DevInfo->Vendor)	    Found->Vendor = DevInfo->Vendor;
	if (DevInfo->Model)	    Found->Model = DevInfo->Model;
	if (DevInfo->ModelDesc)	    Found->ModelDesc = DevInfo->ModelDesc;
	if (DevInfo->Ident)	    Found->Ident = DevInfo->Ident;
	if (DevInfo->Serial)	    Found->Serial = DevInfo->Serial;
	if (DevInfo->Capacity)	    Found->Capacity = DevInfo->Capacity;
	if (DevInfo->Speed)	    Found->Speed = DevInfo->Speed;
	if (DevInfo->Revision)	    Found->Revision = DevInfo->Revision;
	if (DevInfo->DescList)	    Found->DescList = DevInfo->DescList;

	return 1;
    }

    return 0;
}

/*
 * Get (or create) master for DevInfo
 */
static DevInfo_t *GetMaster(DevInfo, TreePtr, SearchNames)
    DevInfo_t 		       *DevInfo;
    DevInfo_t 		      **TreePtr;
    char		      **SearchNames;
{
    static DevFind_t		Find;
    static DevInfo_t	       *Master;

    if (!DevInfo || !TreePtr)
	return (DevInfo_t *) NULL;

    Master = NULL;

#if	DEBUG_DEVADD
    SImsg(SIM_DBG, "DevAdd: GetMaster: Device is <%s>", DevInfo->Name);
#endif

    /*
     * If the device has a master, find the master device.
     * If one doesn't exist in the tree, then add it by recursively
     * calling this function.
     */
    if (DevInfo->Master && !SearchNames) {
	Master = NULL;
	if (*TreePtr) {
	    DevFindInit(&Find);
	    Find.Tree = *TreePtr;
	    Find.NodeName = DevInfo->Master->Name;
	    Find.NodeID = DevInfo->Master->NodeID;
	    Master = DevFind(&Find);

	    if (Master && EQ(Master->Name, DevInfo->Master->Name))
		/* Check and fix any differences in info between master's */
		DevGraft(Master, DevInfo->Master);

	    if (Master && Master != DevInfo->Master) {
		/*
		 * This is normal since callers often create place holder
		 * master devices which are then replaced when the real
		 * device is probed and added.
		 */
#if	0
		SImsg(SIM_DBG,
		      "DevAdd: Found different master with same name <%s> for device <%s>",
		      Master->Name, DevInfo->Name);
#endif

		/*
		 * We don't want extra nearly identical nodes, so reset
		 * the DevInfo's master to be the one already in the Tree.
		 */
		DevInfo->Master = Master;
	    }

	}

	if (!Master) {
	    SImsg(SIM_DBG,
		  "DevAdd: Adding parent <%s> of device <%s> to tree",
		  DevInfo->Master->Name, DevInfo->Name);
#if	DEBUG_DEVADD
	    if (Debug)
		DevDumpPath((TreePtr) ? *TreePtr : NULL, 0, 0);
#endif

	    if (DevAdd(DevInfo->Master, TreePtr, SearchNames) != 0) {
		SImsg(SIM_GERR, "Add master <%s> to device tree failed.", 
		      DevInfo->Name);
		return (DevInfo_t *) NULL;
	    }
	    Master = DevInfo->Master;
	}
    } else {
	if (!*TreePtr) {
	    /*
	     * Must be the root node
	     */
	    *TreePtr = NewDevInfo((DevInfo_t *)NULL);
	    Master = *TreePtr;
	    Master->Model = "<ROOT DEVINFO NODE>";
#if 0	/* XXX Old method */
	    /* Make DevInfo the top node */
	    *TreePtr = DevInfo;
#endif
	}
	/* 
	 * The device doesn't have a master, so make it a child of the
	 * top most level.
	 */
	Master = *TreePtr;
    }

    return Master;
}

/*
 * --RECURSE--
 * Add a device to a device list.
 */
extern int DevAdd(DevInfo, TreePtr, SearchNames)
    DevInfo_t 		       *DevInfo;
    DevInfo_t 		      **TreePtr;
    char		      **SearchNames;
{
    register DevInfo_t 	       *Master = NULL, *mp = NULL;
    static DevFind_t		Find;
    int				Okay;
    int				Status;

    if (!DevInfo || !TreePtr) {
	SImsg(SIM_GERR, "Invalid parameter passed to AddDevice()");
	return(-1);
    }

    SImsg(SIM_DBG, "DevAdd: Adding device <%s>",
	  PRTS(DevInfo->Name));

    /*
     * Only add the device if it's in our search list
     */
    if (SearchNames) {
	Okay = 0;
	if (DevInfo->Name && SearchCheck(DevInfo->Name, SearchNames))
	    Okay = 1;
	if (DevInfo->AltName && SearchCheck(DevInfo->AltName, SearchNames))
	    Okay = 1;
	if (!Okay) {
	    SImsg(SIM_DBG, "DevAdd: Device <%s> not in search list.",
		  DevInfo->Name);
	    return(0);	/* Lie */
	}
    }

    Status = CheckExists(DevInfo, TreePtr);
    if (Status < 0)		/* Error */
	return -1;
    else if (Status > 0)	/* Already exists */
	return 0;

    if (DevInfo->Name)
	DevInfo->Name = strdup(DevInfo->Name);		/* Leak */

    Master = GetMaster(DevInfo, TreePtr, SearchNames);
    if (!Master) {
	SImsg(SIM_DBG, "DevAdd: Cannot find/create master for device <%s>",
	      DevInfo->Name);
	return -1;
    }

    /*
     * Perform device specific sanitizing of DevInfo
     */
    switch (DevInfo->Type) {
    case DT_DISKDRIVE:		DevDiskDriveSanitize(DevInfo);	break;
    }

    /*
     * Add it provided it's not the root node itself
     */
    if (Master) {
	if (Master->Name)
	    Master->Name = strdup(Master->Name);	/* Leak */

	if (Master->Slaves) {
	    /* Add to existing list of slaves */
	    for (mp = Master->Slaves; mp && mp->Next; mp = mp->Next)
		if (mp == mp->Next) {
		    SImsg(SIM_DBG, "DevAdd: Loop detected? Node=<%s>",
			  PRTS(mp->Name));
		    return -1;
		}
	    mp->Next = DevInfo;
	} else {
	    /* Create first slave */
	    Master->Slaves = DevInfo;
	}
    }

    return(0);
}

/*
 * Backwards compatibility
 */
extern int AddDevice(DevInfo, TreePtr, SearchNames)
    DevInfo_t 		       *DevInfo;
    DevInfo_t 		      **TreePtr;
    char		      **SearchNames;
{
    return DevAdd(DevInfo, TreePtr, SearchNames);
}
