/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.11 $";
#endif

#include "defs.h"
#include "dmi.h"

typedef struct {
    char		       *Desc;		/* Description */
    int				Type;		/* Device Type */
    int				Class;		/* Class Type */
} ConnType_t;

/*
 * Connector Types (Section 3.3.9.2)
 */
static ConnType_t ConnTypes[] = {
    { "" },
    { "Centronics" },
    { "Mini-Centronics" },
    { "Proprietary" },
    { "DB-25M" },
    { "DB-25F" },
    { "DB-15M" },
    { "DB-15F" },
    { "DB-9M" },
    { "DB-9F" },
    { "RJ-11" },
    { "RJ-45" },
    { "50-pin Mini-SCSI" },
    { "Mini-DIN" },
    { "Micro-DIN" },
    { "PS/2" },
    { "Infrared" },
    { "HP-HIL" },
    { "USB", DT_SERIAL, CT_USB },
    { "SSA SCSI" },
    { "Circular DIN-8M" },
    { "Circular DIN-8F" },
    { "On-board IDE", DT_CONTROLLER, CT_ATA },
    { "On-board Floppy", DT_FLOPPY, 0 },
    { "9-pin Dual Inline (pin 10 cut)" },
    { "25-pin Dual Inline (pin 25 cut)" },
    { "50-pin Dual Inline" },
    { "68-pin Dual Inline" },
    { "On-board CD-ROM Sound Input" },
    { "Mini-Centronics Type-14" },
    { "Mini-Centronics Type-26" },
    { "Mini-jack (headphones)" },
    { "BNC" },
    { "1394", DT_BUS, CT_FIREWIRE },
    { "PC-98" },
    { "PC-98Hires" },
    { "PC-H98" },
    { "PC-98Note" },
    { "PC-98Full" },
    { NULL },
};
#define CONNTYPES_SIZE	sizeof(ConnTypes)/sizeof(ConnType_t *)

/*
 * Fetch a ConnType
 */
static ConnType_t *GetConnType(Entry)
     int			Entry;
{
    if (Entry > CONNTYPES_SIZE || !ConnTypes[Entry].Desc ||
	ConnTypes[Entry].Desc[0] == CNULL)
	return (ConnType_t *) NULL;

    return &ConnTypes[Entry];
}

/*
 * Probe a parallel port to see if it has anything attached
 */
static int DmiParPortProbe(DevInfo)
     DevInfo_t		       *DevInfo;
{
    DevInfo_t		       *New;
    static ProbeData_t		ProbeData;
    static DevData_t		DevData;

    if (!DevInfo)
	return -1;

#if	defined(HAVE_PROBEPARPORT)
    ProbeDataInit(&ProbeData);
    DevDataInit(&DevData);
    DevData.DevUnit = DevInfo->Unit;
    ProbeData.DevData = &DevData;

    if (New = ProbeParPort(&ProbeData))
	DevInfo->Slaves = New;
#endif	/* HAVE_PROBEPARPORT */

    return 0;
}

/*
 * Decode DMI Port Information (DMI_STRUCT_PORT)
 */
extern DevInfo_t *DmiDecodePort(Hdr, Data, Dmi, Query)
     DmiHeader_t	       *Hdr;
     void		       *Data;
     Dmi_t		       *Dmi;
     DmiQuery_t		       *Query;
{
    DevInfo_t		       *DevInfo;
    DevDesc_t		      **lp;
    DmiStructPort_t	       *Port;
    ConnType_t		       *ConnType;
    char		       *Internal = NULL;
    char		       *External = NULL;
    char		       *cp;
    char		       *Desc = NULL;
    char		       *ModelDesc = NULL;
    char		       *Dev = NULL;
    static char			Buff[128];

    Port = (DmiStructPort_t *) Data;

    if (cp = DmiString(Hdr, Port->IntLocation))
	Internal = strdup(cp);
    if (cp = DmiString(Hdr, Port->ExtLocation))
	External = strdup(cp);
    SImsg(SIM_DBG, 
	  "DMI PORT: Handle=0x%04X Type=0x%02X Internal=<%s> Ext=<%s>",
	  Hdr->Handle, Port->Type, PRTS(Internal), PRTS(External));
    SImsg(SIM_DBG, 
	  "DMI PORT:\tsizeof(DmiStructPort_t)=%d (0x%X) Hdr->Length=%d (0x%X)",
	  sizeof(DmiStructPort_t), sizeof(DmiStructPort_t), 
	  Hdr->Length, Hdr->Length);

    /*
     * If this is a specific query, then set the values now and return
     */
    if (Query && Query->Type == Hdr->Type) {
	/* Return nothing for now */
	return DevInfo;
    }

    DevInfo = NewDevInfo(NULL);
    DevInfo->Type = DT_GENERIC;
    lp = &(DevInfo->DescList);

    if (ConnType = GetConnType(Port->IntConn)) {
	if (ConnType->Type)
	    DevInfo->Type = ConnType->Type;
	if (ConnType->Class)
	    DevInfo->ClassType = ConnType->Class;
    }
    if (ConnType = GetConnType(Port->ExtConn)) {
	if (ConnType->Type)
	    DevInfo->Type = ConnType->Type;
	if (ConnType->Class)
	    DevInfo->ClassType = ConnType->Class;
    }

    switch (Port->Type) {
	/* Values per section 3.3.9.3 */
    case 0x00:
	break;
    case 0x01:	Desc = "Parallel Port XT/AT Compatable";
    case 0x02:	Desc = "Parallel Port PS/2";
    case 0x03:	Desc = "Parallel Port ECP";
    case 0x04:	Desc = "Parallel Port EPP";
    case 0x05:	Desc = "Parallel Port ECP/EPP";
	DevInfo->Type = DT_PARALLEL;
	Dev = "parallel";
	DmiParPortProbe(DevInfo);
	break;
    case 0x06:	Desc = "Serial Port XT/AT Compatable";
    case 0x07:	Desc = "Serial Port 16450 Compatable";
    case 0x08:	Desc = "Serial Port 16550 Compatable";
    case 0x09:	Desc = "Serial Port 16550A Compatable";
	DevInfo->Type = DT_SERIAL;
	Dev = "com";
	break;
    case 0x0A:
	SImsg(SIM_DBG, "DMI PORT: Ignoring SCSI device");
	return (DevInfo_t *) NULL;
#if 0
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_SCSI;
	Dev = "scsi";
#endif
	break;
    case 0x0B:
	DevInfo->Type = DT_AUDIO;
	Desc = "MIDI";
	Dev = "midi";
	break;
    case 0x0C:
	Desc = "Joy Stick";
	Dev = "joystick";
	break;
    case 0x0D:
	DevInfo->Type = DT_KEYBOARD;
	Dev = "kbd";
	break;
    case 0x0E:
	DevInfo->Type = DT_POINTER;
	Dev = "mouse";
	break;
    case 0x0F:
	SImsg(SIM_DBG, "DMI PORT: Ignoring SCSI device");
	return (DevInfo_t *) NULL;
#if 0
	Desc = "SSA";
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_SCSI;
	Dev = "scsi";
#endif
	break;
    case 0x10:
	DevInfo->Type = DT_SERIAL;
	DevInfo->ClassType = CT_USB;
	Dev = "usb";
	break;
    case 0x11:
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_FIREWIRE;
	Dev = "firewire";
	break;
    case 0x12:
    case 0x13:	Desc = "PCMCIA Type II";
    case 0x14:	Desc = "PCMCIA Type III";
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_PCMCIA;
	Dev = "pccard";
	break;
    case 0x15:	Desc = "Cardbus";
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_CARDBUS;
	Dev = "cardbus";
	break;
    case 0x16:
	DevInfo->Type = DT_BUS;
	Desc = "Access Bus";
	Dev = "accessbus";
	break;
    case 0x17:
	SImsg(SIM_DBG, "DMI PORT: Ignoring SCSI device");
	return (DevInfo_t *) NULL;
#if 0
	Desc = "SCSI-II";
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_SCSI;
	Dev = "scsi";
#endif
	break;
    case 0x18:
	SImsg(SIM_DBG, "DMI PORT: Ignoring SCSI device");
	return (DevInfo_t *) NULL;
#if 0
	Desc = "Wide";
	DevInfo->Type = DT_BUS;
	DevInfo->ClassType = CT_SCSI;
	Dev = "scsi";
#endif
	break;
    case 0x19:
	Desc = "PC-98";
	break;
    case 0x1A:
	Desc = "PC-98-Hires";
	break;
    case 0x1B:
	Desc = "PC-H98";
	break;
    case 0x1C:
	SImsg(SIM_DBG, "DMI PORT: Ignoring framebuffer device");
	return (DevInfo_t *) NULL;
#if 0
	DevInfo->Type = DT_FRAMEBUFFER;
	Dev = "fb";
#endif
	break;
    case 0x1D:
	SImsg(SIM_DBG, "DMI PORT: Ignoring audio device");
	return (DevInfo_t *) NULL;
#if 0
	DevInfo->Type = DT_AUDIO;
	Dev = "audio";
#endif
	break;
    case 0x1E:
	DevInfo->Type = DT_SERIAL;
	DevInfo->ClassType = CT_MODEM;
	Dev = "modem";
	break;
    case 0x1F:
	SImsg(SIM_DBG, "DMI PORT: Ignoring network interface device");
	return (DevInfo_t *) NULL;
#if 0
	DevInfo->Type = DT_NETIF;
	Dev = "eth";
#endif
	break;
    case 0xA0:
	Desc = "8251 Compatible";
	break;
    case 0xA1:
	Desc = "8251 FIFO Compatible";
	break;
    default:
	SImsg(SIM_DBG, "DMI PORT: Unknown Port Type=0x%02X", Port->Type);
    }

    if (Desc)
	ModelDesc = strdup(Desc);
    else if (External)
	ModelDesc = External;
    else if (Internal)
	ModelDesc = Internal;

    /*
     * Set any special values according to Type
     */
    switch (DevInfo->Type) {
    case DT_FLOPPY:	
	Dev = "fd";
	if (EQ(ModelDesc, "FLOPPY"))
	    ModelDesc = NULL;
	break;
    }

    /*
     * Set any special values according to class
     */
    switch (DevInfo->ClassType) {
    case CT_ATA:	Dev = "ide";		break;
    case CT_USB:
	if (EQN(External, "USB", 3))
	    ModelDesc = NULL;
	break;
    }

    if (!Dev)
	Dev = "port";
    DevInfo->Unit = UnitAssign(0, 0, Dev);
    snprintf(Buff, sizeof(Buff), "%s%d", Dev, DevInfo->Unit);
    DevInfo->Name = strdup(Buff);

    DevInfo->ModelDesc = ModelDesc;

    if (External)
	AddDesc(lp, "External", External, 0);
    if (Internal)
	AddDesc(lp, "Internal", Internal, 0);

    if (ConnType = GetConnType(Port->IntConn))
	AddDesc(lp, "Internal Connector", strdup(ConnType->Desc), 0);
    if (ConnType = GetConnType(Port->ExtConn))
	AddDesc(lp, "External Connector", strdup(ConnType->Desc), 0);

    return DevInfo;
}
