/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.52 $";
#endif

/*
 * Common SysInfo SCSI functions
 */

#include "defs.h"
#include "myscsi.h"

static char *CDLoadTypes[] 	= CD_LOAD_TYPES;
#define MAX_CD_LOAD_TYPES 	sizeof(CDLoadTypes)/sizeof(char *)

/*
 * SCSI Command Names
 */
static ScsiCommand_t ScsiCommands[] = {
    { SCSI_INQUIRY, 	0,				"INQUIRY" }, 
    { SCSI_INQUIRY, 	SCSI_INQUIRY_IDENT, 		"IDENT" }, 
    { SCSI_INQUIRY, 	SCSI_INQUIRY_SERIAL,		"SERIAL" },
    { SCSI_CAPACITY, 	0,				"CAPACITY" },
    { SCSI_MODE_SENSE, 	0,				"MODE SENSE" },
    { SCSI_MODE_SENSE, 	SCSI_MODE_CD_CAP,		"CD CAP" },
    { SCSI_MODE_SENSE, 	SCSI_DAD_MODE_FORMAT,		"DAD FORMAT" },
    { SCSI_MODE_SENSE, 	SCSI_DAD_MODE_SPEED, 		"DAD SPEED" },
    { SCSI_MODE_SENSE, 	SCSI_DAD_MODE_GEOMETRY,		"DAD GEOMETRY" },
    { SCSI_MODE_SENSE, 	SCSI_SAD_MODE_DATA_COMP,	"SAD DATA COMP" },
};
#define ScsiCommandsSize (sizeof(ScsiCommands)/sizeof(ScsiCommand_t))

/*
 * Specs:
 *   1X CD-ROM speed: 176KB/s (original audio CD)
 *   1X DVD-ROM speed: ~ 1.3MB/s (about 9X CD speed)
 * Drive specs for Sun DVD drives:
 *   Toshiba SD-M1401 DVD: 10X/40X (up to 13,500KB/s DVD, 6843KB/s CD)
 */
#define CD_SPEED_1X		176	/* KB/s */
#define DVD_SPEED_1X		1300	/* KB/s */

/*
 * Convert a raw SCSI byte stream into printable ASCII.
 * Used for debug output.
 */
extern char *ScsiRawDump(Data, DataLen)
     char		       *Data;
     int			DataLen;
{
    static char			String[1024];
    char			Buf[8];
    register int		i;

    for (String[0] = CNULL, i = 0; i < DataLen; ++i) {
	snprintf(Buf, sizeof(Buf), "%s0x%2.2x",
		 (i == 0) ? "" : " ", (char)Data[i]);
	StrAppend(String, sizeof(String), Buf);
    }

    return String;
}

/*
 * Get SCSI Command definetion
 */
extern ScsiCommand_t *ScsiGetCommand(Cmd, PageCode)
     int			Cmd;
     int			PageCode;
{
    register int		i;

    for (i = 0; i < ScsiCommandsSize; ++i)
	if (ScsiCommands[i].Cmd == Cmd && ScsiCommands[i].PageCode == PageCode)
	    return &ScsiCommands[i];

    return (ScsiCommand_t *) NULL;
}

/*
 * Get SCSI Command Name
 */
extern char *ScsiGetCommandName(Cmd, PageCode)
     int			Cmd;
     int			PageCode;
{
    ScsiCommand_t	       *sc;

    if (sc = ScsiGetCommand(Cmd, PageCode))
	return sc->Name;

    return "";	/* Return something safe to print */
}

/*
 * Is SCSI command (Query->PageCode) supported by the current SCSI device?
 */
static int ScsiSupCmd(Query)
     ScsiQuery_t	       *Query;
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;
    static ScsiSupCmd_t		SupCmd;
    ScsiCdbG0_t		       *QueryCdb;
    char		       *DevName = NULL;
    char		       *Status = NULL;
    char		       *Label = "SCSI SUP CMD";
    int				Supported = 0;
    register int		i;
    char		       *CmdName;

   if (!Query) {
	SImsg(SIM_DBG, "ScsiSupCmd: Bad parameters");
	return 0;
    }

    if (Query->DevInfo && Query->DevInfo->Name && Query->DevInfo->Name[0])
	DevName = Query->DevInfo->Name;
    else
	DevName = Query->DevFile;

    CmdName = ScsiGetCommandName(Query->Cmd, Query->PageCode);
    QueryCdb = (ScsiCdbG0_t *) Query->Cdb;

    SImsg(SIM_DBG, "\t%s: %s: Name=<%s> Cmd=0x%2.2x PageCode=0x%2.2x",
	  PRTS(DevName), Label, CmdName, Query->Cmd, Query->PageCode);

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = SCSI_INQUIRY;
    /*
     * If EVPD is set, then we want to check to see if 
     * Query->PageCode is supported.  We do this by asking
     * the device for a list of all the EVPD pagecodes it supports,
     * then checking that list to for Query->PageCode.
     *
     * If EVPD is not set, then we want to ask the device if the
     * command Query->Cmd is supported.
     */
    if (QueryCdb->EVPD) {
	/*
	 * Retrieve the Supported EVPD pagecodes
	 */
	Cdb.EVPD = 1;
	Cdb.CmdDt = 0;
    } else {
	/*
	 * Retrieve list of supported commands
	 */
	Cdb.EVPD = 0;
	Cdb.CmdDt = 1;
    }
    Cdb.Length = (Word_t) sizeof(ScsiSupCmd_t);

    if (Debug)
	SImsg(SIM_DBG, "\t%s: %s: Send: %s",
	      PRTS(DevName), Label, ScsiRawDump((char *)&Cdb, sizeof(Cdb)));

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    Supported = 0;
    if (ScsiCmd(&Cmd, Query) == 0) {
	if (Debug)
	    SImsg(SIM_DBG, "\t%s: %s: Results: %s",
		  PRTS(DevName), Label, ScsiRawDump((char *)Cmd.Data, 12));
	    
	/*
	 * Decode results
	 */
	memcpy(&SupCmd, Cmd.Data, sizeof(SupCmd));
	if (QueryCdb->EVPD) {
	    /* Supported EVPD pagecode? */
	    Status = "UNSUPPORTED EVPD PageCode";
	    /*
	     * If Query->PageCode is found in Data within Size bytes,
	     * then it's supported by the device.
	     */
	    for (i = 0;
		 i < SupCmd.Evpd.Size && i < sizeof(SupCmd.Evpd.Data);
		 ++i) {
		if (SupCmd.Evpd.Data[i] == Query->PageCode) {
		    Status = "SUPPORTED EVPD";
		    Supported = 1;
		    break;
		}
	    }
	} else {
	    /* Supported command? */
	    /*
	     * Only an explicit value of Support == UNSUPPORTED will result in
	     * returning 0 for this command.  The assumption being that older
	     * devices which don't properly implement this query will return
	     * different values for Support.  Newer, compliant devices which
	     * return an explicit UNSUPPORTED are taken at their word.
	     */
	    Supported = 1;
	    Status = NULL;
	    switch (SupCmd.CmdDt.Support) {
	    case 0:	/* 000b */
		Status = "Supported Command Query unsupported";
		break;
	    case 1:	/* 001b */
		Status = "UNSUPPORTED CmdDt";
		Supported = 0;
		break;
	    case 2:	/* 010b */
		Status = "Reserved 010b";
		break;
	    case 3:	/* 011b */
		Status = "SUPPORTED CmdDt";
		break;
	    case 4:	/* 100b */
		Status = "Vendor specific status";
		break;
	    case 5:	/* 101b */
		Status = "Supported in vendor specific manner";
		break;
	    case 6:	/* 110b */
		Status = "Vendor specific status";
		break;
	    case 7:	/* 111b */
		Status = "Reserved 111b";
		break;
	    }
	}

	SImsg(SIM_DBG, 
	      "\t%s: %s: Is \"%s\" supported? Status=%s",
	      PRTS(DevName), Label, CmdName, PRTS(Status));
    } else {
	/*
	 * XXX Assume that devices which don't support the query for
	 * supported commands do indeed support this particuliar command.
	 */
	Supported = 1;
	SImsg(SIM_DBG, 
	      "\t%s: %s: Check for command \"%s\" support failed",
	      PRTS(DevName), Label, CmdName);
    }
    
    return Supported;
}

/*
 * Decode a SCSI CAPACITY command and add info to DevInfo as we can.
 */
extern int ScsiCapacityDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiCapacity_t	       *Capacity;
    DiskDrive_t		       *Disk;
    DevInfo_t		       *DevInfo;

    if (!Query|| !Query->DevInfo || !Query->Data)
	return(-1);

    Capacity = (ScsiCapacity_t *) Query->Data;
    DevInfo = Query->DevInfo;

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: Blocks=<%ld> BlockSize=<%d>",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  ntohl(Capacity->Blocks), 
	  ntohl(Capacity->BlkSize));

    if (!(Disk = DiskSetup(DevInfo, "ScsiCapacityDecode")))
	return(-1);

    /*
     * Set what we know
     */
    Disk->Size = nsect_to_mbytes( ntohl(Capacity->Blocks),
				  ntohl(Capacity->BlkSize) );
    Disk->SecSize = ntohl(Capacity->BlkSize);

    return(0);
}

/*
 * Decode results from SCSI_INQUIRY_IDENT
 * An Identifier Descriptor is suppose to uniquely identify the device
 * with the purpose of allowing for the detection of multi-pathed devices.
 *
 * This is a SCSI-3 feature which many SCSI devices do not yet implement.
 */
extern int ScsiIdentDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiIdent_t		       *ScsiIdent;
    Ident_t		       *Ident;
    DevInfo_t		       *DevInfo;

    if (!Query || !Query->Data || !Query->DevInfo) {
	SImsg(SIM_DBG, "ScsiIdentDecode: Bad parameters");
	return(-1);
    }

    ScsiIdent = (ScsiIdent_t *) Query->Data;
    DevInfo = Query->DevInfo;

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: PageCode=0x%02x CodeSet=0x%x Type=%d IdLen=%d",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  ScsiIdent->Hdr.PageCode, 
	  ScsiIdent->Descriptors[0].CodeSet, 
	  ScsiIdent->Descriptors[0].IdentType,
	  ScsiIdent->Descriptors[0].IdentLength	  );
    if (ScsiIdent->Descriptors[0].IdentLength > 0)
	SImsg(SIM_DBG, 
	      "\t%s: SCSI %s: ID=<%.*s>",
	      DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	      ScsiIdent->Descriptors[0].IdentLength,
	      ScsiIdent->Descriptors[0].Identifier);

    /*
     * We only look at the 1st descriptor.
     * Officially this data is only valid if:
     *
     *		PageCode==0x83	(SCSI_INQUIRY_IDENT)
     *		CodeSet > 0 && CodeSet << 0xF
     *		IdentType > 0 && IdentType << 0xF
     *
     * Many older (pre SCSI-3) devices will not understand the
     * Cdb.EVPD and Cdb.Addr1=SCSI_INQUIRY_IDENT bits and will
     * return normal INQUIRY data.
     */
#define InRange(v)	( (v) > 0 && (v) < 0xF )
    if (ScsiIdent->Hdr.PageCode == SCSI_INQUIRY_IDENT &&
	InRange(ScsiIdent->Descriptors[0].CodeSet) &&
	InRange(ScsiIdent->Descriptors[0].IdentType) &&
	ScsiIdent->Descriptors[0].IdentLength > 0) {

	Ident = IdentCreate(NULL);
	Ident->Length = ScsiIdent->Descriptors[0].IdentLength;
	if (ScsiIdent->Descriptors[0].CodeSet == IDCS_ASCII)
	    Ident->Type = IDT_ASCII;
	else if (ScsiIdent->Descriptors[0].CodeSet == IDCS_BINARY)
	    Ident->Type = IDT_BINARY;
	/* Length + CNULL */
	Ident->Identifier = (u_char *) xmalloc(Ident->Length + 1);

	(void) copybytes(ScsiIdent->Descriptors[0].Identifier,
			 sizeof(ScsiIdent->Descriptors[0].Identifier),
			 Ident->Identifier, Ident->Length);

	/* Terminate with NULL regardless of type */
	Ident->Identifier[Ident->Length] = CNULL;

	DevInfo->Ident = Ident;
    } else {
	SImsg(SIM_DBG, "\t%s: SCSI %s: Information Invalid - ignored",
	      DevInfo->Name,  ScsiGetCommandName(Query->Cmd, Query->PageCode));
    }

#undef	InRange

    return 0;
}

/*
 * Decode results from SCSI_INQUIRY_SERIAL
 * This value represents the product's serial number as assigned by the
 * vendor.
 *
 * This is a SCSI-3 feature which many SCSI devices do not yet implement.
 */
extern int ScsiSerialDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiSerialNum_t	       *Serial;
    DevInfo_t		       *DevInfo;
    char		       *cp;

    if (!Query || !Query->Data || !Query->DevInfo) {
	SImsg(SIM_DBG, "ScsiSerialDecode: Bad parameters");
	return(-1);
    }

    Serial = (ScsiSerialNum_t *) Query->Data;
    DevInfo = Query->DevInfo;

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: PageCode=0x%02x Length=%d Serial=<%.*s>",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  Serial->Hdr.PageCode, 
	  Serial->Hdr.Length, 
	  Serial->Hdr.Length, 
	  Serial->Serial);

    /*
     * We only look at the 1st descriptor.
     * Officially this data is only valid if:
     *
     *		PageCode==0x83	(SCSI_INQUIRY_SERIAL)
     *
     * Many older (pre SCSI-3) devices will not understand the
     * Cdb.EVPD and Cdb.Addr1=SCSI_INQUIRY_SERIAL bits and will
     * return normal INQUIRY data.
     */
    if (Serial->Hdr.PageCode == SCSI_INQUIRY_SERIAL &&
	Serial->Hdr.Length > 0) {
	if (cp = CleanString(Serial->Serial, 
			     sizeof(Serial->Hdr.Length), 
			     CSF_ENDNONPR|CSF_ALPHANUM)) {
	    /* Serial must be at least 6 chars long to be valid */
	    if (strlen(cp) >= 6)
		DevInfo->Serial = strdup(cp);
	    else
		SImsg(SIM_DBG, "\t%s: Serial value <%s> too short.  Ignored.",
		      DevInfo->Name, cp);
	
	}
    } else {
	SImsg(SIM_DBG, "\t%s: SCSI %s: Information Invalid - ignored",
	      DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode));
    }

    return 0;
}

/*
 * Decode a SCSI INQUIRY command and add info to DevInfo as we can.
 */
extern int ScsiInquiryDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiInquiry_t	       *Inq;
    DevInfo_t		       *DevInfo;
    Define_t		       *Def;
    char		       *cp;
    int				IsEMC = 0;
    ScsiInquiryEMCProtection_t *EMCProt;

    if (!Query || !Query->Data || !Query->DevInfo) {
	SImsg(SIM_DBG, "ScsiInquiryDecode: Bad parameters");
	return(-1);
    }

    Inq = (ScsiInquiry_t *) Query->Data;
    DevInfo = Query->DevInfo;

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: vendor=<%.*s> product=<%.*s> type=0x%02x",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  sizeof(Inq->VendorID), Inq->VendorID, 
	  sizeof(Inq->ProductID), Inq->ProductID, Inq->DevType);
    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s:    rev=<%.*s> serial=<%.*s> vendorspec=<%.*s>",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  sizeof(Inq->Revision), Inq->Revision, 
	  sizeof(Inq->Serial), Inq->Serial,
	  sizeof(Inq->_vendor3), Inq->_vendor3
	  );

    /*
     * This is implied by this function being called
     */
    DevInfo->ClassType = CT_SCSI;

    /*
     * Decode results
     */
#define IsOk(s)	( s && s[0] && s[0] != ' ' && isalnum(s[0]) )
    if (IsOk(Inq->VendorID))
	if (cp = CleanString(Inq->VendorID, sizeof(Inq->VendorID), 0))
	    DevInfo->Vendor = strdup(cp);
    if (IsOk(Inq->ProductID))
	if (cp = CleanString(Inq->ProductID, sizeof(Inq->ProductID), 0))
	    DevInfo->Model = strdup(cp);
    if (IsOk(Inq->Revision))
	if (cp = CleanString(Inq->Revision, sizeof(Inq->Revision), 0))
	    DevInfo->Revision = strdup(cp);

    /*
     * EMC specific checks
     */
    if (EQ(DevInfo->Vendor, "EMC") && EQ(DevInfo->Model, "SYMMETRIX")) {
	IsEMC = 1;
	EMCProt = (ScsiInquiryEMCProtection_t *) & (Inq->Serial[8]);
	AddDevDesc(DevInfo, "True", "EMC Disk", DA_APPEND);
    }

    /*
     * Serial number decoding
     */
    if (IsOk(Inq->Serial))
	/* EMC Serial numbers are only 8 chars long */
	if (cp = CleanString(Inq->Serial, 
			     IsEMC ? 8 : sizeof(Inq->Serial), 
			     CSF_ENDNONPR|CSF_ALPHANUM)) {
	    /* Serial must be at least 6 chars long to be valid */
	    if (strlen(cp) >= 6)
		DevInfo->Serial = strdup(cp);
	    else
		SImsg(SIM_DBG, "%s: Serial value <%s> too short - ignored",
		      DevInfo->Name, cp);
	}
#undef IsOk

    /*
     * What SCSI version are we?
     */
    cp = NULL;
    switch ((int)Inq->RDF) {
    case RDF_LEVEL0:	cp = "SCSI-1";	break;
    case RDF_SCSI2:	cp = "SCSI-2";	break;
    case RDF_CCS:	cp = "CCS";	break;
    }
    if (cp)
	AddDevDesc(DevInfo, cp, "SCSI Version/Protocol", DA_APPEND);
    else
	SImsg(SIM_UNKN, "Unknown RDF/SCSI Version 0x%x", 
	      Inq->RDF);

    /*
     * ANSI/ECMA/ISO version info
     */
    cp = NULL;
    Def = DefGet(DL_SCSI_ANSI, (char *) NULL, (long) Inq->ANSI, 0);
    if (Def && Def->ValStr2)
	cp = Def->ValStr2;
    else {
	SImsg(SIM_DBG, "%s: Unknown %s=0x%02x", 
	      DevInfo->Name, DL_SCSI_ANSI, Inq->ANSI);
	cp = itoa(Inq->ANSI);
    }
    AddDevDesc(DevInfo, cp, "ANSI Version", DA_APPEND);
    AddDevDesc(DevInfo, itoa(Inq->ECMA), "ECMA Version", DA_APPEND);
    AddDevDesc(DevInfo, itoa(Inq->ISO), "ISO Version", DA_APPEND);

    /*
     * Check all the flag bits
     */
    if (Inq->WBus16)
	AddDevDesc(DevInfo, "Wide SCSI: 16-bit Data Transfers", 
		   "Supports", DA_APPEND);
    if (Inq->WBus32)
	AddDevDesc(DevInfo, "Wide SCSI: 32-bit Data Transfers", 
		   "Supports", DA_APPEND);
    if (!Inq->WBus16 && !Inq->WBus32)
	AddDevDesc(DevInfo, "Narrow SCSI: 8-bit Data Transfers", 
		   "Supports", DA_APPEND);
    if (Inq->Addr16)
	AddDevDesc(DevInfo, "Wide SCSI: 16-bit Addressing", 
		   "Supports", DA_APPEND);
    if (Inq->Addr32)
	AddDevDesc(DevInfo, "Wide SCSI: 32-bit Addressing", 
		   "Supports", DA_APPEND);
    if (!Inq->Addr16 && !Inq->Addr32)
	AddDevDesc(DevInfo, "Narrow SCSI: 8-bit Addressing", 
		   "Supports", DA_APPEND);

    if (Inq->ACKQREQQ)
	AddDevDesc(DevInfo, "Data Transfer on Secondary Bus", 
		   "Supports", DA_APPEND);

    if (Inq->Sync)
	AddDevDesc(DevInfo, "Syncronous Data Transfers", "Supports", 
		   DA_APPEND);

    if (Inq->Removable)
	AddDevDesc(DevInfo, "Removable Media", "Supports", DA_APPEND);

    if (Inq->NormACA)
	AddDevDesc(DevInfo, "Normal ACA (NACA)", "Supports", DA_APPEND);
    if (Inq->TrmTsk)
	AddDevDesc(DevInfo, "TERMINATE TASK", "Supports", DA_APPEND);
    if (Inq->AERC)
	AddDevDesc(DevInfo, "Async Event Notification", "Supports", DA_APPEND);
    if (Inq->EncServ)
	AddDevDesc(DevInfo, "Enclosure Services", "Supports", DA_APPEND);
    if (Inq->MultiPort)
	AddDevDesc(DevInfo, "Multi-Ported Device", "Supports", DA_APPEND);
    if (Inq->MChngr)
	AddDevDesc(DevInfo, "Embedded/attached to medium changer", 
		   "Supports", DA_APPEND);
    if (Inq->TranDis)
	AddDevDesc(DevInfo, "Transfer Disable Messages", "Supports", 
		   DA_APPEND);
    if (Inq->CmdQue)
	AddDevDesc(DevInfo, "Command Queuing", "Supports", DA_APPEND);
    if (Inq->Linked)
	AddDevDesc(DevInfo, "Linked Commands", "Supports", DA_APPEND);
    if (Inq->RelAddr)
	AddDevDesc(DevInfo, "Relative Addressing", "Supports", DA_APPEND);

    if (IsEMC) {
	if (EMCProt->Mirror)
	    AddDevDesc(DevInfo, "EMC Mirrored", "Supports", DA_APPEND); 
	if (EMCProt->RAID)
	    AddDevDesc(DevInfo, "EMC RAID-S", "Supports", DA_APPEND); 
	if (EMCProt->RDFEnabled) {
	    if (EMCProt->PriRDFDev)
	        AddDevDesc(DevInfo, "EMC RDF R1", "Supports", DA_APPEND); 
	    else
	        AddDevDesc(DevInfo, "EMC RDF R2", "Supports", DA_APPEND); 
	}
    }

    /*
     * Look up device type
     */
    Def = DefGet(DL_SCSI_DTYPE, (char *) NULL, (long) Inq->DevType, 0);
    if (Def) {
	/*
	 * Set Device Type 
	 */
	if (Def->ValStr1) {
	    DevType_t		*dtPtr;

	    dtPtr = TypeGetByName(Def->ValStr1);
	    if (dtPtr)
		DevInfo->Type = dtPtr->Type;
	    else
		SImsg(SIM_DBG, "%s: SCSIdtype DevType=<%s> is unknown.",
		      DevInfo->Name, Def->ValStr1);
	}
    } else
	SImsg(SIM_DBG, "%s: Unknown SCSIdtype=0x%02x", 
	      DevInfo->Name, Inq->DevType);

    return(0);
}
     
/*
 * Get SCSI Device Identification
 */
extern int ScsiQueryIdent(Query)
     ScsiQuery_t	       *Query;
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiQueryIdent: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_INQUIRY;
    Cdb.Addr1 = Query->PageCode = SCSI_INQUIRY_IDENT;
    Cdb.EVPD = 1;		/* Special - Enable Vital Product Data */
    Cdb.Length = (Word_t) sizeof(ScsiIdent_t);
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Query->Data = Cmd.Data;
	return ScsiIdentDecode(Query);
    } else
	return(-1);
}

/*
 * Get SCSI Device Serial Number
 */
extern int ScsiQuerySerial(Query)
     ScsiQuery_t	       *Query;
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiQuerySerial: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_INQUIRY;
    Cdb.Addr1 = Query->PageCode = SCSI_INQUIRY_SERIAL;
    Cdb.EVPD = 1;		/* Special - Enable Vital Product Data */
    Cdb.Length = (Word_t) sizeof(ScsiIdent_t);
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Query->Data = Cmd.Data;
	return ScsiSerialDecode(Query);
    } else
	return(-1);
}

/*
 * Get description info about SCSI devices by sending a SCSI INQUIRY
 * to the device.
 */
extern int ScsiQueryInquiry(Query)
     ScsiQuery_t	       *Query;
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiInquiry: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_INQUIRY;
    Query->PageCode = 0;
    Cdb.Length = sizeof(ScsiInquiry_t);

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Query->Data = Cmd.Data;
	return ScsiInquiryDecode(Query);
    } else
	return(-1);
}

/*
 * Get description info about SCSI devices by sending a SCSI CAPACITY
 * to the device.
 */
extern int ScsiQueryCapacity(Query)
     ScsiQuery_t	       *Query;	
{
    static ScsiCdbG1_t		Cdb;
    static ScsiCmd_t		Cmd;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiCapacity: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_CAPACITY;
    Query->PageCode = 0;
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Query->Data = Cmd.Data;
	return ScsiCapacityDecode(Query);
    } else
	return(-1);
}

/*
 * Decode a SCSI GEOMETRY command and add info to DevInfo as we can.
 */
extern int ScsiGeometryDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiGeometry_t	       *Geometry;
    DiskDrive_t		       *Disk;
    DevInfo_t		       *DevInfo;
    u_int			PhyCyl;

    if (!Query|| !Query->DevInfo || !Query->Data)
	return(-1);

    Geometry = (ScsiGeometry_t *) Query->Data;
    DevInfo = Query->DevInfo;
    PhyCyl =  word3_to_uint(Geometry->NumCyls);

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: #tracks=<%d> PhyCyl=<%d> RPM=<%d>",
	  DevInfo->Name,  ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  Geometry->Heads, PhyCyl, ntohs(Geometry->RPM));

    if (!(Disk = DiskSetup(DevInfo, "ScsiGeometryDecode")))
	return(-1);

    Disk->Tracks = Geometry->Heads;
    Disk->PhyCyl = PhyCyl;
    Disk->RPM = ntohs(Geometry->RPM);

    return(0);
}

/*
 * Decode a SCSI FORMAT command and add info to DevInfo as we can.
 */
extern int ScsiFormatDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiFormat_t	       *Format;
    DiskDrive_t		       *Disk;
    DevInfo_t		       *DevInfo;

    if (!Query|| !Query->DevInfo || !Query->Data)
	return(-1);

    Format = (ScsiFormat_t *) Query->Data;
    DevInfo = Query->DevInfo;

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: #sect=<%d> secsize=<%d>",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  ntohs(Format->SectTrack), 
	  ntohs(Format->DataBytesSect));

    if (!(Disk = DiskSetup(DevInfo, "ScsiFormatDecode")))
	return(-1);

    Disk->Sect = ntohs(Format->SectTrack);
    Disk->SecSize = ntohs(Format->DataBytesSect);
    Disk->Tracks = ntohs(Format->TracksPerZone);
    Disk->AltSectPerZone = ntohs(Format->AltSectZone);
    Disk->AltTracksPerVol = ntohs(Format->AltTracksVol);
    Disk->AltTracksPerZone = ntohs(Format->AltTracksZone);
    Disk->IntrLv = ntohs(Format->Interleave);
    Disk->TrackSkew = ntohs(Format->TrackSkew);
    Disk->CylSkew = ntohs(Format->CylinderSkew);

    return(0);
}

/*
 * Get description info about SCSI devices by sending a SCSI FORMAT
 * to the device.
 */
extern int ScsiQueryFormat(Query)
     ScsiQuery_t	       *Query;	
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;
    static ScsiFormat_t		Format;
    static ScsiModePage_t	ModePage;
    ScsiModeHeader_t	       *Hdr;
    char		       *LocPtr;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiFormat: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_MODE_SENSE;
    Cdb.Addr1 = Query->PageCode = SCSI_DAD_MODE_FORMAT;
    Cdb.Length = 255;	/* XXX Yuck.  It's hardcoded */
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Hdr = (ScsiModeHeader_t *) Cmd.Data;
	LocPtr = SCSI_MODE_PAGE_ADDR(Hdr, char);
	(void) memset(&Format, 0, sizeof(Format));
	/* 
	 * First grab the ModePage header which contains the value
	 * for ModePage.Length which indicates the size of the actual
	 * data to copy.
	 */
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t),
			 &ModePage, sizeof(ModePage));
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t) + ModePage.Length, 
			 &Format, sizeof(Format));
	Query->Data = (void *) &Format;
	return ScsiFormatDecode(Query);
    } else
	return(-1);
}

/*
 * Decode a SCSI SPEED command and add info to DevInfo as we can.
 */
extern int ScsiSpeedDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiSpeed_t	 	       *Speed;
    DiskDrive_t		       *Disk;
    DevInfo_t		       *DevInfo;

    if (!Query|| !Query->DevInfo || !Query->Data)
	return(-1);

    Speed = (ScsiSpeed_t *) Query->Data;
    DevInfo = Query->DevInfo;

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: Speed=%d",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  ntohs(Speed->Speed));

    return(0);
}

/*
 * Get description info about SCSI devices by sending a SCSI SPEED
 * to the device.
 */
extern int ScsiQuerySpeed(Query)
     ScsiQuery_t	       *Query;	
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;
    static ScsiSpeed_t		Speed;
    static ScsiModePage_t	ModePage;
    ScsiModeHeader_t	       *Hdr;
    char		       *LocPtr;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiSpeed: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_MODE_SENSE;
    Cdb.Addr1 = Query->PageCode = SCSI_DAD_MODE_SPEED;
    Cdb.Length = 255;	/* XXX Yuck.  It's hardcoded */
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Hdr = (ScsiModeHeader_t *) Cmd.Data;
	LocPtr = SCSI_MODE_PAGE_ADDR(Hdr, char);
	(void) memset(&Speed, 0, sizeof(Speed));
	/* 
	 * First grab the ModePage header which contains the value
	 * for ModePage.Length which indicates the size of the actual
	 * data to copy.
	 */
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t),
			 &ModePage, sizeof(ModePage));
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t) + ModePage.Length, 
			 &Speed, sizeof(Speed));
	Query->Data = (void *) &Speed;
	return ScsiSpeedDecode(Query);
    } else
	return(-1);
}

/*
 * Get description info about SCSI devices by sending a SCSI GEOMETRY
 * to the device.
 */
extern int ScsiQueryGeometry(Query)
     ScsiQuery_t	       *Query;	
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;
    static ScsiGeometry_t	Geometry;
    static ScsiModePage_t	ModePage;
    ScsiModeHeader_t	       *Hdr;
    char		       *LocPtr;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiGeometry: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_MODE_SENSE;
    Cdb.Addr1 = Query->PageCode = SCSI_DAD_MODE_GEOMETRY;
    Cdb.Length = 255;	/* Yuck.  It's hardcoded */
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Hdr = (ScsiModeHeader_t *) Cmd.Data;
	LocPtr = SCSI_MODE_PAGE_ADDR(Hdr, char);
	(void) memset(&Geometry, 0, sizeof(Geometry));
	/* 
	 * First grab the ModePage header which contains the value
	 * for ModePage.Length which indicates the size of the actual
	 * data to copy.
	 */
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t),
			 &ModePage, sizeof(ModePage));
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t) + ModePage.Length, 
			 &Geometry, sizeof(Geometry));
	Query->Data = (void *) &Geometry;
	return ScsiGeometryDecode(Query);
    } else
	return(-1);
}

/*
 * Decode a SCSI CD Capability command and add info to DevInfo as we can.
 */
extern int ScsiCDCapDecode(Query)
    ScsiQuery_t		       *Query;
{
    static char			Buff[64];
    ScsiCDCap_t		       *Cap;
    DevInfo_t		       *DevInfo;
    Desc_t		      **d;
    int				CDspeedX = 0;
    int				DVDspeedX = 0;
    int				MMCversion = 0;
    int				PageLen = 0;
    int				CapValid = 1;
    char		       *CmdName;

    if (!Query|| !Query->DevInfo || !Query->Data)
	return(-1);

    CmdName = ScsiGetCommandName(Query->Cmd, Query->PageCode);
    DevInfo = Query->DevInfo;
    d = &DevInfo->DescList;

    Cap = (ScsiCDCap_t *) Query->Data;
    if (!Cap) {
	SImsg(SIM_DBG, "\t%s: SCSI %s: Data invalid - no Capability data",
	      DevInfo->Name, CmdName);
	return -1;
    }
    PageLen = (int) Cap->ModePage.Length;
    if (PageLen < 0x14) {
	SImsg(SIM_DBG, "\t%s: SCSI %s: unrecognized Capability ModePage Length 0x%X",
	      DevInfo->Name, CmdName, PageLen);
	CapValid = 0;
    }

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: CurRdSpd=%d MaxRdSpd=%d CurWrSpd=%d MaxWrSpd=%d DVD-ROMread=%d",
	  DevInfo->Name, CmdName,
	  ntohs(Cap->CurReadSpeed), ntohs(Cap->MaxReadSpeed),
	  ntohs(Cap->CurWriteSpeed), ntohs(Cap->MaxWriteSpeed),
	  Cap->DVD_ROM_Read);
    /*
     * ModePage.Length indicates the amount of info provided.
     * It also indicates which revision of the SCSI MMC the returned
     * data represents.
     */
    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: ModePage Length=%d (0x%X)",
	  DevInfo->Name, CmdName, PageLen, PageLen);
    switch (PageLen) {
    case 0x14:	/* MMC-1 spec CD Capabilities page length */
	SImsg(SIM_DBG, "\t%s: SCSI %s: using MMC-1", DevInfo->Name, CmdName);
	MMCversion = 1;
	break;
    case 0x16:	/* MMC-2 bytes 24-25 are reserved, but they might be omitted */
    case 0x17:
    case 0x18:	/* MMC-2 spec CD Capabilities page length */
	SImsg(SIM_DBG, "\t%s: SCSI %s: using MMC-2", DevInfo->Name, CmdName);
	MMCversion = 2;
	break;
    default:
	SImsg(SIM_DBG, "\t%s: SCSI %s: unrecognized Capability ModePage Length 0x%X",
	      DevInfo->Name, CmdName, PageLen);
	/* Keep going because the data provided may be backwards compatible */
	break;
    }
    if (MMCversion) {
	snprintf(Buff, sizeof(Buff), "MMC-%d", MMCversion);
	AddDesc(d, "Supports", Buff, 0);
    }

    /* Check to see if Capability data looks valid */
    if (Cap->MaxReadSpeed && ntohs(Cap->MaxReadSpeed)) {
	CDspeedX = (int) ntohs(Cap->MaxReadSpeed) / CD_SPEED_1X;
	if (CDspeedX < 1 || CDspeedX > 1000) {
	    /* Bogus speed ranges */
	    SImsg(SIM_DBG, "\t%s: SCSI %s: Invalid - CD Speed %dX",
		  DevInfo->Name, CmdName, CDspeedX);
	    CapValid = 0;
	}
    }
    if (Cap->CurReadSpeed && Cap->MaxReadSpeed &&
	ntohs(Cap->CurReadSpeed) > ntohs(Cap->MaxReadSpeed)) {
	SImsg(SIM_DBG, 
	      "\t%s: SCSI %s: Invalid - Current Read Speed > Maximum",
	      DevInfo->Name, CmdName);
	CapValid = 0;
    }
    if (Cap->CurWriteSpeed && Cap->MaxWriteSpeed &&
	ntohs(Cap->CurWriteSpeed) > ntohs(Cap->MaxWriteSpeed)) {
	SImsg(SIM_DBG, 
	      "\t%s: SCSI %s: Invalid - Current Write Speed > Maximum",
	      DevInfo->Name, CmdName);
	CapValid = 0;
    }
    if (Cap->MaxWriteSpeed && Cap->MaxReadSpeed &&
	ntohs(Cap->MaxWriteSpeed) > ntohs(Cap->MaxReadSpeed)) {
	SImsg(SIM_DBG, 
	      "\t%s: SCSI %s: Invalid - Max Write Speed > Max Read Speed",
	      DevInfo->Name, CmdName);
	CapValid = 0;
    }
    if (Cap->DVD_R_Read && !Cap->DVD_ROM_Read) {
	SImsg(SIM_DBG, 
	      "\t%s: SCSI %s: Invalid - Device can read DVD-R, but not DVD-ROM",
	      DevInfo->Name, CmdName);
	CapValid = 0;
    }
    if (Cap->DVD_RAM_Read && !Cap->DVD_ROM_Read) {
	SImsg(SIM_DBG, 
	      "\t%s: SCSI %s: Invalid - Device can read DVD-RAM, but not DVD-ROM",
	      DevInfo->Name, CmdName);
	CapValid = 0;
    }
    if (!CapValid) {
	SImsg(SIM_DBG,
	      "\t%s: SCSI %s: Ignoring CD Capability data since it looks invalid",
	      DevInfo->Name, CmdName);
	return -1;
    }

    if (Cap->DVD_ROM_Read) {
	/* If we can read DVD-ROM, then let's set our type now */
	DevInfo->Type = DT_DVD;
    }

    if (Cap->MaxReadSpeed && ntohs(Cap->MaxReadSpeed)) {
	/*
	 * If this is a DVD, it attempts to report the DVD read speed
	 * in addition to the CD-ROM read speed, i.e. "4X/24X"
	 * All > 1X speed drives say "up to nnX" instead of "nnX".
	 */
	if (DevInfo->Type == DT_DVD)
	    DVDspeedX = (int) ntohs(Cap->MaxReadSpeed) / DVD_SPEED_1X;

	Buff[0] = CNULL;
	if (CDspeedX && DVDspeedX)
	    (void) snprintf(Buff, sizeof(Buff), "%dX/%dX", 
			    DVDspeedX, CDspeedX);
	else if (CDspeedX)
	    (void) snprintf(Buff, sizeof(Buff), "%dX",  CDspeedX);

	if (Buff[0])
	    DevInfo->Speed = strdup(Buff);
	if (DevInfo->Type == DT_DVD && DVDspeedX) {
	    (void) snprintf(Buff, sizeof(Buff), "up to %dX", DVDspeedX);
	    AddDesc(d, "DVD Speed", Buff, 0);
	}
	(void) snprintf(Buff, sizeof(Buff), "up to %dX", CDspeedX);
	AddDesc(d, "CD Speed", Buff, 0);
    }

    /*
     * Media Support
     */
    /* Assume everything can read CD-ROM */
    AddDesc(d, "Reads Media", "CD-ROM", 0);
    if (Cap->CD_R_Read) AddDesc(d, "Reads Media", "CD-R", 0);
    if (Cap->CD_RW_Read) AddDesc(d, "Reads Media", "CD-RW", 0);
    if (Cap->DVD_ROM_Read) AddDesc(d, "Reads Media", "DVD-ROM", 0);
    if (Cap->DVD_R_Read) AddDesc(d, "Reads Media", "DVD-R", 0);
    if (Cap->DVD_RAM_Read) AddDesc(d, "Reads Media", "DVD-RAM", 0);
    if (Cap->Mode2Form1) AddDesc(d, "Reads Media", "Mode-2 Form 1 (XA)", 0);
    if (Cap->Mode2Form2) AddDesc(d, "Reads Media", "Mode-2 Form 2", 0);
    if (Cap->MultiSession) AddDesc(d, "Reads Media", "Multi-session", 0);
    if (Cap->CD_DA_Supported) AddDesc(d, "Reads Media", "Digital Audio", 0);

    if (Cap->CD_R_Write) AddDesc(d, "Writes Media", "CD-R", 0);
    if (Cap->CD_RW_Write) AddDesc(d, "Writes Media", "CD-RW", 0);
    if (Cap->DVD_R_Write) AddDesc(d, "Writes Media", "DVD-R", 0);
    if (Cap->DVD_RAM_Write) AddDesc(d, "Writes Media", "DVD-RAM", 0);

    /*
     * Misc
     */
    if (Cap->TestWrite) AddDesc(d, "Supports", "Emulation/Test Write", 0);
    if (Cap->AudioPlay) AddDesc(d, "Supports", "Audio Play", 0);
    if (Cap->Composite) AddDesc(d, "Supports", "Composite A/V stream", 0);
    if (Cap->DigitalPort1) 
        AddDesc(d, "Supports", "Digital Output on Port 1", 0);
    if (Cap->DigitalPort2) 
        AddDesc(d, "Supports", "Digital Output on Port 2", 0);
    if (Cap->CD_DA_Accurate) 
        AddDesc(d, "Supports", "Digital Audio Stream Accurate", 0);
    if (Cap->RW_Supported) 
        AddDesc(d, "Supports", "R-W Sub Channel Info", 0);
    if (Cap->RW_DeintCorr) 
        AddDesc(d, "Supports", "R-W de-interleaved Sub Channel", 0);
    if (Cap->RW_LeadIn)
	AddDesc(d, "Supports", "Read R-W channel info from Lead-In", 0);
    if (Cap->C2_Pointers) AddDesc(d, "Supports", "C2 Error Pointers", 0);
    if (Cap->ISRC) AddDesc(d, "Supports", "ISRC information", 0);
    if (Cap->UPC) AddDesc(d, "Supports", "UPC Catalog Number", 0);
    if (Cap->ReadBarCode) AddDesc(d, "Supports", "Read Bar Codes", 0);
    if (Cap->BufUnder) 
	AddDesc(d, "Supports", "Buffer under-run free recording", 0);

    if (Cap->LoadingType) {
	if (Cap->LoadingType < MAX_CD_LOAD_TYPES)
	    AddDesc(d, "Loading Type", CDLoadTypes[Cap->LoadingType], 0);
	else
	    AddDesc(d, "Loading Type", itoa(Cap->LoadingType), 0);
    }

    if (Cap->SepChanVol) 
	AddDesc(d, "Supports", "Volume controls each channel seperately", 0);
    if (Cap->SepChanMute) 
	AddDesc(d, "Supports", "Mute controls each channel seperately", 0);
    if (Cap->DiskPresentRep) 
	AddDesc(d, "Supports", "Changer disk present reporting", 0);
    if (Cap->SWSlotSel)
	AddDesc(d, "Supports", "Changer slot is software selectable", 0);
    if (Cap->SideChange)
	AddDesc(d, "Supports", "Changer can select both sides of disc", 0);
    if (Cap->CopyMgtRev) 
	AddDesc(d, "Supports", "Copy Management Revision", 0);

    if (Cap->NumVolLevels)
	AddDesc(d, "Number Volume Levels", 
		itoa((int)ntohs(Cap->NumVolLevels)), 0);

    if (Cap->BufferSize)
	AddDesc(d, "Buffer Size", 
		GetSizeStr((Large_t)ntohs(Cap->BufferSize), KBYTES), 0);

    if (Cap->MaxReadSpeed) 
	AddDesc(d, "Max Read Speed per sec", 
		GetSizeStr((Large_t)ntohs(Cap->MaxReadSpeed), KBYTES), 0);
    if (Cap->CurReadSpeed) 
	AddDesc(d, "Current Read Speed per sec", 
		GetSizeStr((Large_t)ntohs(Cap->CurReadSpeed), KBYTES), 0);
    if (Cap->MaxWriteSpeed) 
	AddDesc(d, "Max Write Speed per sec", 
		GetSizeStr((Large_t)ntohs(Cap->MaxWriteSpeed), KBYTES), 0);
    if (Cap->CurWriteSpeed) 
	AddDesc(d, "Current Write Speed per sec", 
		GetSizeStr((Large_t)ntohs(Cap->CurWriteSpeed), KBYTES), 0);

    return(0);
}

/*
 * Decode a SCSI Data Compression command and add info to DevInfo as we can.
 */
extern int ScsiDataCompDecode(Query)
    ScsiQuery_t		       *Query;
{
    ScsiDataComp_t	       *Comp;
    DevInfo_t		       *DevInfo;
    Define_t		       *Def;
    u_int			CompAlg = 0;
    u_int			DeCompAlg = 0;

    if (!Query|| !Query->DevInfo || !Query->Data)
	return(-1);

    Comp = (ScsiDataComp_t *) Query->Data;
    DevInfo = Query->DevInfo;
    CompAlg = ntohl(Comp->CompAlg);
    DeCompAlg = ntohl(Comp->DecompAlg);

    SImsg(SIM_DBG, 
	  "\t%s: SCSI %s: Comp Algo=<0x%02x> Decomp Algo=<0x%02x>",
	  DevInfo->Name, ScsiGetCommandName(Query->Cmd, Query->PageCode),
	  CompAlg, DeCompAlg);

    /*
     * Look up algorithm type
     */
    Def = DefGet(DL_SCSI_COMP_ALG, (char *) NULL, (long) CompAlg, 0);
    if (Def && Def->ValStr2)
	AddDesc(&DevInfo->DescList, "Compression Algorithm", Def->ValStr2, 0);
    else
	SImsg(SIM_DBG, "%s: Unknown %s=0x%02x", 
	      DevInfo->Name, DL_SCSI_COMP_ALG, CompAlg);

    Def = DefGet(DL_SCSI_COMP_ALG, (char *) NULL, (long) DeCompAlg, 0);
    if (Def && Def->ValStr2)
	AddDesc(&DevInfo->DescList, "Decompression Algorithm", 
		Def->ValStr2, 0);
    else
	SImsg(SIM_DBG, "%s: Unknown %s=0x%02x", 
	      DevInfo->Name, DL_SCSI_COMP_ALG, DeCompAlg);

    return(0);
}

/*
 * Get the CD Capabilities of a SCSI device.
 * This only works on CD/DVD based devices
 */
extern int ScsiQueryCDCap(Query)
     ScsiQuery_t	       *Query;	
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;
    static ScsiCDCap_t		Cap;
    static ScsiModePage_t	ModePage;
    ScsiModeHeader_t	       *Hdr;
    char		       *LocPtr;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiQueryCDCap: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_MODE_SENSE;
    Cdb.Addr1 = Query->PageCode = SCSI_MODE_CD_CAP;
    Cdb.Length = 255;	/* XXX Yuck.  It's hardcoded */
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Hdr = (ScsiModeHeader_t *) Cmd.Data;
	LocPtr = SCSI_MODE_PAGE_ADDR(Hdr, char);
	(void) memset(&Cap, 0, sizeof(Cap));
	/* 
	 * First grab the ModePage header which contains the value
	 * for ModePage.Length which indicates the size of the actual
	 * data to copy.
	 */
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t),
			 &ModePage, sizeof(ModePage));
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t) + ModePage.Length, 
			 &Cap, sizeof(Cap));
	Query->Data = (void *) &Cap;
	return ScsiCDCapDecode(Query);
    } else
	return(-1);
}

/*
 * Get the Data Compression info for a Sequential Access device.
 */
extern int ScsiQueryDataComp(Query)
     ScsiQuery_t	       *Query;	
{
    static ScsiCdbG0_t		Cdb;
    static ScsiCmd_t		Cmd;
    static ScsiDataComp_t	Comp; 
    static ScsiModePage_t	ModePage;
    ScsiModeHeader_t	       *Hdr;
    char		       *LocPtr;

    if (!Query) {
	SImsg(SIM_DBG, "ScsiQueryDataComp: Bad parameters");
	return(-1);
    }

    /*
     * Initialize
     */
    memset(&Cdb, 0, sizeof(Cdb));
    Cdb.Cmd = Query->Cmd = SCSI_MODE_SENSE;
    Cdb.Addr1 = Query->PageCode = SCSI_SAD_MODE_DATA_COMP;
    Cdb.Length = 255;	/* XXX Yuck.  It's hardcoded */
    Query->Cdb = (void *) &Cdb;

    memset(&Cmd, 0, sizeof(Cmd));
    Cmd.Cdb = &Cdb;
    Cmd.CdbLen = sizeof(Cdb);
    Cmd.DevFD = Query->DevFD;
    Cmd.DevFile = Query->DevFile;

    if (!ScsiSupCmd(Query))
	return -1;

    if (ScsiCmd(&Cmd, Query) == 0) {
	/*
	 * Decode results
	 */
	Hdr = (ScsiModeHeader_t *) Cmd.Data;
	LocPtr = SCSI_MODE_PAGE_ADDR(Hdr, char);
	(void) memset(&Comp, 0, sizeof(Comp));
	/* 
	 * First grab the ModePage header which contains the value
	 * for ModePage.Length which indicates the size of the actual
	 * data to copy.
	 */
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t),
			 &ModePage, sizeof(ModePage));
	(void) copybytes(LocPtr, sizeof(ScsiModePage_t) + ModePage.Length, 
			 &Comp, sizeof(Comp));
	Query->Data = (void *) &Comp;
	return ScsiDataCompDecode(Query);
    } else
	return(-1);
}

/*
 * Query a SCSI device
 */
extern int ScsiQuery(DevInfo, DevFile, DevFD, OverRide)
    DevInfo_t		       *DevInfo;
    char		       *DevFile;
    int				DevFD;
    int				OverRide;
{
    int				Status = 0;
    int				Type = -1;
    static ScsiQuery_t		Query;
    ScsiInquiry_t	       *Inq;
    int				fd = -1;

    if (!DevInfo || !DevFile) {
	SImsg(SIM_DBG, "ScsiQuery: Bad parameters.");
	return(-1);
    }

    if (DevFD < 0) {
	/*
	 * We use O_RDWR because many OS's require Write perms
	 */
	DevFD = fd = open(DevFile, O_RDWR|O_NDELAY|O_NONBLOCK);
	if (fd < 0) {
	    SImsg(SIM_GERR, "%s: ScsiQuery: open for read failed: %s",
		  DevFile, SYSERR);
	    return(-1);
	}
    }

    (void) memset(&Query, 0, sizeof(Query));
    Query.DevInfo = DevInfo;
    Query.DevFile = DevFile;
    Query.DevFD = DevFD;
    Query.OverRide = OverRide;

    /* Insure DevInfo->Name is valid */
    if (!Query.DevInfo->Name)
	Query.DevInfo->Name = DevFile;

    /*
     * Do a general inquiry on the device.
     * After this we use the Device Type for further commands.
     */
    if (ScsiQueryInquiry(&Query) == 0) {
	++Status;
	Inq = (ScsiInquiry_t *) Query.Data;
	Type = Inq->DevType;
    }

    if (ScsiQueryIdent(&Query) == 0)
	++Status;
    if (ScsiQuerySerial(&Query) == 0)
	++Status;

    /*
     * Perform device type specific queries.
     */
    switch (Type) {
    case SCSI_DTYPE_DAD:
	/* Direct Access Devices (disk drives, etc.) */
	if (ScsiQueryCapacity(&Query) == 0)
	    ++Status;
	if (ScsiQueryFormat(&Query) == 0)
	    ++Status;
	if (ScsiQueryGeometry(&Query) == 0)
	    ++Status;
	if (ScsiQuerySpeed(&Query) == 0)
	    ++Status;
	break;
    case SCSI_DTYPE_CDROM:
	/* CD/DVD */
	if (ScsiQueryCDCap(&Query) == 0)
	    ++Status;
	break;
    case SCSI_DTYPE_SEQ:
	/* Sequential Access devices (tape drives, etc.) */
	if (ScsiQueryDataComp(&Query) == 0)
	    ++Status;
	break;
    }

    /*
     * If we opened the descriptor, then close it now
     */
    if (fd >= 0)
	(void) close(fd);

    /*
     * If at least 1 function succeeded, we return success.
     */
    if (Status > 0)
	return 0;
    else
	return -1;
}
