/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.23 $";
#endif

/*
 * System Configuration
 *
 * Use the sysconf() routine to retrieve system configuration info.
 * This routine is available on POSIX compliant systems.
 */
#include "defs.h"
#include "getsysconf.h"

/*
 * Show System Configuration information utilizing the sysconf()
 * system call.
 *
 * Returns ptr to Define_t in Query->Out:
 * 	KeyStr		SysConf Variable (string)
 *	ValStr1		Type of variable (Boolean, etc)
 *	ValStr2		Description of Variable
 *	ValStr3		Value (string)
 *	ValInt1		Value (integer)
 */
extern int GetSysConf(Query)
     MCSIquery_t	       *Query;
{
#if	defined(HAVE_SYSCONF)
    long			Value;
    int				NumValid = 0;
    char		       *ValuePtr;
    Define_t		       *SysConfDef = NULL;
    Define_t		       *Def;
    Define_t		       *Last = NULL;
    SysConf_t		       *sc;

    if (!Query) {
	errno = ENXIO;
	return -1;
    }

    if (Query->Op == MCSIOP_DESTROY)
	return DefDestroy((Define_t *) Query->Out);
    /* Else do MCSIOP_CREATE */

    for (sc = &SysConfTable[0]; sc && sc->Desc; ++sc) {
	/*
	 * Call sysconf().  If it succeeds, then store the return value
	 * in ValInt1.  If it fails, set ValInt1 to -1 and continue on.
	 */
	Value = sysconf(sc->SysConfVar);
	if (Value < 0) {
	    SImsg(SIM_GERR, "sysconf(%s) failed: %s", sc->VarName, SYSERR);
	    continue;
	}
	/* Create a Define_t variable for this entry */
	Def = (Define_t *) xcalloc(1, sizeof(Define_t));
	Def->ValInt1 = Value;
	Def->KeyStr = sc->VarName;
	Def->ValStr2 = sc->Desc;

	switch (sc->Type) {
	case SCT_BOOL:
	    Def->ValStr1 = "bool";
	    if (Value)
		ValuePtr = "TRUE";
	    else
		ValuePtr = "FALSE";
	    break;
	case SCT_LONG:
	    Def->ValStr1 = "long";
	    ValuePtr = strdup(itoa(Value));
	    break;
	}
	/* Store string value here */
	Def->ValStr3 = ValuePtr;

	/* Add to linked list */
	if (Last) {
	    Last->Next = Def;
	    Last = Def;
	} else {
	    SysConfDef = Last = Def;
	}

	++NumValid;
    }

    Query->Out = (Opaque_t) SysConfDef;
    Query->OutSize = NumValid;
#endif	/* HAVE_SYSCONF */

    return 0;
}
