/*
 * Copyright (c) 1992-2002 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/4.2/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.20 $";
#endif

/*
 * Setup and installation program
 */

#if	defined(HAVE_AUTOCONFIG_H)
#include "autoconfig.h"
#endif	/* HAVE_AUTOCONFIG_H */
#if	defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif
#include <stdio.h>
#include "options.h"
#include "mcl.h"
#include "version.h"
#include "defs.h"

int				DebugFlag = 0;
int				AcceptLicense = 0;
int				ConfirmFlag = 0;
int				OverRide = 0;
int				DoPrintVersion = 0;
int				IgnoreErrors = 0;
int				OpInstall = 0;
int				OpLicense = 0;
int				OpUnInstall = 0;
int				OpVerify = 0;
int				OpInventory = 0;
int				TermWidth = 80;
char			       *McdMapFile = NULL;
char			       *McdInfoFile = NULL;
char			       *InstallDir = NULL;
char			       *LicenseFile = NULL;
char			       *Answer = NULL;
char			       *Answers[8];
char			       *AnswersYN[3] = { A_YES, A_NO, NULL };
char			       *MCLprogramName = "sysinfo";
static Meter_t			MeterInfo;

/*
 * Command line options table.
 */
OptionDescRec Opts[] = {
    {"--acceptlicense", NoArg,	OptBool,(OptPtr_t) &AcceptLicense,	"1",
     	NULL,  "Accept the terms of the license without prompting" },
    {"--debug", NoArg,	OptBool,	(OptPtr_t) &DebugFlag,		"1",
     	NULL,  "Enable debugging" },
    {"--confirm", NoArg,OptBool,	(OptPtr_t) &ConfirmFlag,	"1",
     	NULL,  "Confirm all default answers (no prompts)" },
    {"--ignoreerrors", NoArg,OptBool,	(OptPtr_t) &IgnoreErrors,	"1",
     	NULL,  "Ignore normally fatal errors" },
    {"--install", NoArg,OptBool,	(OptPtr_t) &OpInstall,		"1",
     	NULL,  "Install software" },
    {"--installdir", SepArg,OptStr,	(OptPtr_t) &InstallDir,		NULL,
     	NULL,  "Installation directory" },
    {"--inventory", NoArg,OptBool,	(OptPtr_t) &OpInventory,	"1",
     	NULL,  "Inventory distribution files" },
    {"--license", NoArg,OptBool,	(OptPtr_t) &OpLicense,		"1",
     	NULL,  "Install license" },
    {"--licensefile", SepArg,OptStr,	(OptPtr_t) &LicenseFile,        NULL,
     	NULL,  "Name of license file to be installed" },
    {"--mcdmap", SepArg,OptStr,		(OptPtr_t) &McdMapFile,		MCDMAP,
     	NULL,  "MCD Map File" },
    {"--mcdinfo", SepArg,OptStr,	(OptPtr_t) &McdInfoFile,        NULL,
     	NULL,  "MCD Info File" },
    {"--override", NoArg,OptBool,	(OptPtr_t) &OverRide,		"1",
     	NULL,  "Override errors which prompt for confirmation" },
    {"--uninstall", NoArg,OptBool,	(OptPtr_t) &OpUnInstall,	"1",
     	NULL,  "UnInstall software" },
    {"--verify", NoArg,OptBool,		(OptPtr_t) &OpVerify,		"1",
     	NULL,  "Verify installed software" },
    {"--version",NoArg,	OptBool,	(OptPtr_t) &DoPrintVersion,	"1",
	 NULL,	"Print version of this program" },
};

/*
 * Get program version
 */
static char *GetVersion()
{
    char		       *cp;
    char		       *Version = NULL;
    static char			Buff[128];
    static char			String[128];

    if (String[0])
	return String;

    strncpy(Buff, RCSid, sizeof(Buff));
    if (cp = strchr(Buff, ' ')) {
	Version = cp+1;
	if (cp = strchr(Version, ' '))
	    *cp = CNULL;
    }

    if (Version) {
	(void) snprintf(String, sizeof(String), "%s version %s",
			ProgramName, Version);
	return String;
    }

    return (char *) NULL;
}

/*
 * Bail out for some kind of fatal condition
 */
extern void Bail()
{
    if (!IgnoreErrors) {
	Msg(0, "\nThe installation has been aborted due to an error.\n");
	exit(1);
    }
}

/*
 * Bail out at user request
 */
extern void UserBail()
{
    Msg(0, "\nThe installation has been aborted at your request.\n");
    exit(1);
}

/*
 * Ask a question
 */
extern char *Ask(Question, AnswerList, Default)
     char		       *Question;
     char		      **AnswerList;
     char		       *Default;
{
    char			Prompt[512];
    char		       *Ans = NULL;
    size_t			Len;
    char		      **cpp;
    int				i;

    if (!Question)
	return (char *) NULL;

    snprintf(Prompt, sizeof(Prompt), "%s ", Question);
    Len = strlen(Prompt);

    if (AnswerList) {
	snprintf(&Prompt[Len], sizeof(Prompt)-Len, "(");
	Len = strlen(Prompt);
    }
	
    for (cpp = AnswerList; cpp && *cpp; ++cpp) {
	if (cpp != AnswerList) {
	    snprintf(&Prompt[Len], sizeof(Prompt)-Len, ",");
	    ++Len;
	}
	snprintf(&Prompt[Len], sizeof(Prompt)-Len, "%s", *cpp);
	Len = strlen(Prompt);
    }
    if (AnswerList) {
	snprintf(&Prompt[Len], sizeof(Prompt)-Len, ") ");
	Len = strlen(Prompt);
    } else {
	snprintf(&Prompt[Len], sizeof(Prompt)-Len, " ");
	Len = strlen(Prompt);
    }

    if (Default) {
	snprintf(&Prompt[Len], sizeof(Prompt)-Len, "[%s] ", Default);
	Len = strlen(Prompt);
    }

    for (i = 0; i < SETUP_MAX_PROMPTS; ++i) {
	Msg(0, "%s", Prompt);

	if (ConfirmFlag && Default) {
	    Ans = Default;
	    Msg(0, "%s\n", Ans);
	    return Ans;
	}

	Ans = GetInput(0);
	Len = strlen(Ans);
	if (Default && Len == 0)
	    return Default;
	if (Len == 0)
	    continue;

	if (!AnswerList)
	    return Ans;

	for (cpp = AnswerList; cpp && *cpp; ++cpp)
	    if (EQ(Ans, *cpp))
		return Ans;
	Msg(0, "The input \"%s\" is not valid.\n", Ans);
    }

    Error("Maximum prompt limit reached.");
    return (char *) NULL;
}

/*
 * Get pathname relative to distribution directory
 */
extern char *GetDistPath(File, DistDir)
     char		       *File;
     char		       *DistDir;
{
    static char			Path[MAXPATHLEN];

    if (FileExists(File))
	snprintf(Path, sizeof(Path), "%s", File);
    else if (DistDir)
	snprintf(Path, sizeof(Path), "%s/files/%s", DistDir, File);
    else
	snprintf(Path, sizeof(Path), "files/%s", File);

    return Path;
}

/*
 * Get an MCD data file
 */
extern char *GetMcdFile(File, Base, DistDir)
     char		       *File;
     char		       *Base;
     char		       *DistDir;
{
    static char		        Path[MAXPATHLEN];

    if (File) {
	if (FileExists(File))
	    return File;
	else {
	    Error("%s: File does not exist.", File);
	    return (char *) NULL;
	}
    } else if (FileExists(Base)) {
	return Base;
    } else if (DistDir) {
	snprintf(Path, sizeof(Path), "%s/%s", DistDir, Base);
	if (FileExists(Path))
	    return Path;

	snprintf(Path, sizeof(Path), "%s/files/%s", DistDir, Base);
	if (FileExists(Path))
	    return Path;
    } else {
	snprintf(Path, sizeof(Path), "files/%s", Base);
	if (FileExists(Path))
	    return Path;
    }

    return (char *) NULL;
}

/*
 * Perform an inventory of the distribution contents
 */
static int Inventory(Mcd, File, DistDir)
     Mcd_t		       *Mcd;
     char		       *File;
     char		       *DistDir;
{
    Mcd_t		       *Entry;
    struct stat			Stat;
    int				ErrCount = 0;
    char		       *Src;

    memset(&MeterInfo, 0, sizeof(MeterInfo));
    MeterInfo.TermWidth = TermWidth;
    MeterInfo.Prompt = "Inventory check . . . ";
    for (Entry = Mcd; Entry; Entry = Entry->Next)
	++(MeterInfo.Total);

    for (Entry = Mcd; Entry; Entry = Entry->Next) {
	++(MeterInfo.Current);
	Meter(&MeterInfo);

	if (!Entry->Src || EQ(Entry->Src, "-"))
	    continue;

	Src = GetDistPath(Entry->Src, DistDir);

	Debug("Checking <%s>", Src);
	if (stat(Src, &Stat) != 0) {
	    Error("%s: Missing file detected", Src);
	    ++ErrCount;
	    continue;
	}
	if (Entry->Size && Stat.st_size != Entry->Size) {
	    Error("%s: File size mismatch - file is %d should be %d", 
		  Src, Stat.st_size, Entry->Size);
	    ++ErrCount;
	    continue;
	}
    }

    if (ErrCount) {
	Msg(0, "\nThere were %d errors during inventory.\n", ErrCount);
	return -1;
    } else {
	Msg(0, "\nInventory succeeded.\n");
	return 0;
    }
}

/*
 * Verify installed distribution is correct
 */
static int Verify(Mcd, Info, DistDir, ProgramPath)
     Mcd_t		       *Mcd;
     McdInfo_t		       *Info;
     char		       *DistDir;
     char		       *ProgramPath;
{
    Mcd_t		       *Entry;
    struct stat			Stat;
    int				ErrCount = 0;
    static char			Path[MAXPATHLEN];

    /*
     * Figure out Installation Directory based on ProgramPath
     */
    SetInstallDir(Info, DistDir, ProgramPath);

    Msg(0, "Verify distribution installed in %s\n", Info->InstallDir);

    memset(&MeterInfo, 0, sizeof(MeterInfo));
    MeterInfo.TermWidth = TermWidth;
    MeterInfo.Prompt = "Verifying . . . ";
    for (Entry = Mcd; Entry; Entry = Entry->Next)
	++(MeterInfo.Total);

    for (Entry = Mcd; Entry; Entry = Entry->Next) {
	++(MeterInfo.Current);
	Meter(&MeterInfo);

	if (!Entry->Src || EQ(Entry->Src, "-"))
	    continue;

	(void) snprintf(Path, sizeof(Path), "%s/%s", 
			Info->InstallDir, Entry->Dest);

	Debug("Checking <%s>", Path);
	if (stat(Path, &Stat) != 0) {
	    Error("%s: Missing file detected", Path);
	    ++ErrCount;
	    continue;
	}
	/*
	 * We can't check size of MCDF_EDIT files because their size will
	 * have changed due to the file edit
	 */
	if (FLAGS_OFF(Entry->Flags, MCDF_EDIT) &&
	    Entry->Size && Stat.st_size != Entry->Size) {
	    Error("%s: File size mismatch - file is %d should be %d", 
		  Path, Stat.st_size, Entry->Size);
	    ++ErrCount;
	    continue;
	}
    }

    if (ErrCount) {
	Msg(0, "\nThere were %d errors during verification.\n", ErrCount);
	return -1;
    } else {
	Msg(0, "\nVerification succeeded.\n");
	return 0;
    }
}

/*
 * Count number of directories found in Path
 */
static int DirCount(Path)
     char		       *Path;
{
    char		       *cp;
    int				Count = 0;

    for (cp = strchr(Path, '/'); cp; cp = strchr(cp+1, '/'))
	++Count;

    return Count;
}

/*
 * qsort directory comparision
 * Sort in descending order
 */
static int DirCmp(p1, p2)
     const void		       *p1;
     const void		       *p2;
{
    char		      **c1 = (char **) p1;
    char		      **c2 = (char **) p2;
    int				l1 = DirCount(*c1);
    int				l2 = DirCount(*c2);

    if (l1 < l2)
	return 1;
    if (l1 > l2)
	return -1;

    return 0;
}

/*
 * Figure out Installation Directory based on ProgramPath
 */
extern int SetInstallDir(Info, DistDir, ProgramPath)
     McdInfo_t		       *Info;
     char		       *DistDir;
     char		       *ProgramPath;
{
    static char			InstallDir[MAXPATHLEN];
    char   		       *cp;

    if (DistDir && IsDir(DistDir))
	(void) snprintf(InstallDir, sizeof(InstallDir), "%s", DistDir);
    else {
	(void) snprintf(InstallDir, sizeof(InstallDir), "%s", ProgramPath);
	if (cp = strrchr(InstallDir, '/')) {
	    *cp = CNULL;
	    if (EQ(InstallDir, ".") || EQN(InstallDir, "./", 2)) {
		if (!getcwd(InstallDir, sizeof(InstallDir))) {
		    Error("Get current working directory failed: %s", SYSERR);
		    return -1;
		}
	    }
	} else {
	    if (!getcwd(InstallDir, sizeof(InstallDir))) {
		Error("Get current working directory failed: %s", SYSERR);
		return -1;
	    }
	}
    }

    Debug("SetInstallDir: InstallDir=<%s>", InstallDir);

    Info->InstallDir = InstallDir;

    return 0;
}

/*
 * Un install distribution
 */
static int UnInstall(Mcd, Info, DistDir, ProgramPath)
     Mcd_t		       *Mcd;
     McdInfo_t		       *Info;
     char		       *DistDir;
     char		       *ProgramPath;
{
    Mcd_t		       *Entry;
    struct stat			Stat;
    int				ErrCount = 0;
    static char			Path[MAXPATHLEN];
    static char			Dir[MAXPATHLEN];
    char		       *cp;
    char		      **Dirs = NULL;
    int				DirCount = 0;
    int				i;

    /*
     * Figure out Installation Directory based on ProgramPath
     */
    SetInstallDir(Info, DistDir, ProgramPath);

    ShowSummary(Info);
    Msg(0, "\n");
    Answer = Ask("Do you want to REMOVE/UNINSTALL this software?", 
		 AnswersYN, A_YES);
    if (!EQ(Answer, A_YES)) {
	Error("You have choosen not to proceed with removing the software.");
	return -1;
    }

    /* Now add support files to the list  of things to remove */
    McdMapAddFile(Mcd, NULL, Info->InstallLic);
    (void) snprintf(Path, sizeof(Path), "%s/%s", 
		    Info->InstallDir, MCDMAP);
    McdMapAddFile(Mcd, NULL, strdup(Path));
    (void) snprintf(Path, sizeof(Path), "%s/%s", 
		    Info->InstallDir, MCDINFO);
    McdMapAddFile(Mcd, NULL, strdup(Path));
    (void) snprintf(Path, sizeof(Path), "%s/%s", 
		    Info->InstallDir, BaseName(ProgramPath));
    McdMapAddFile(Mcd, NULL, strdup(Path));

    /*
     * Setup Meter and count # directories
     */
    memset(&MeterInfo, 0, sizeof(MeterInfo));
    MeterInfo.TermWidth = TermWidth;
    MeterInfo.Prompt = "Removing files . . . ";
    for (Entry = Mcd; Entry; Entry = Entry->Next) {
	++(MeterInfo.Total);
	if (Entry->Type == MCDT_DIR)
	    ++DirCount;
    }

    Dirs = (char **) calloc(DirCount+1, sizeof(char *));
    if (!Dirs) {
	Error("calloc %d bytes failed: %s", sizeof(char *)*DirCount+1);
	Bail();
    }
    DirCount = 0;

    /* Remove all the files and list directories for later deletion */
    for (Entry = Mcd; Entry; Entry = Entry->Next) {
	++(MeterInfo.Current);
	Meter(&MeterInfo);

	if (!Entry->Dest || EQ(Entry->Dest, "-"))
	    continue;

	if (Entry->Dest[0] != '/')
	    (void) snprintf(Path, sizeof(Path), "%s/%s", 
			    Info->InstallDir, Entry->Dest);
	else
	    (void) snprintf(Path, sizeof(Path), "%s", 
			    Entry->Dest);

	Debug("Removing <%s>", Path);
	if (stat(Path, &Stat) != 0) {
	    Error("%s: Missing file detected", Path);
	    ++ErrCount;
	    continue;
	}
	switch (Entry->Type) {
	case MCDT_FILE:
	    if (unlink(Path) != 0) {
		Error("%s: Remove file failed: %s", Path, SYSERR);
		++ErrCount;
	    }
	    /* Try to remove the parent directory without checking */
	    (void) rmdir(DirName(Path));
	    break;
	case MCDT_DIR:
	    Dirs[DirCount++] = strdup(Path);
	    break;
	default:
	    Error("%s: File type %d is unknown.", Path, Entry->Type);
	    ++ErrCount;
	}
    }
    Msg(0, "\n");

    /*
     * Sort Dirs with deepest entries first
     */
    qsort((void *) Dirs, DirCount, sizeof(char **), DirCmp);

    /*
     * Now wack all the Dirs
     */
    memset(&MeterInfo, 0, sizeof(MeterInfo));
    MeterInfo.TermWidth = TermWidth;
    MeterInfo.Prompt = "Removing directories . . . ";
    MeterInfo.Total = DirCount;
    for (i = 0; i < DirCount; ++i) {
	++(MeterInfo.Current);
	Meter(&MeterInfo);

	if (!Dirs[i])
	    break;

	Debug("Removing <%s>", Dirs[i]);
	if (rmdir(Dirs[i]) != 0) {
	    if (errno != ENOENT) {
		Error("%s: Remove directory failed: %s", Dirs[i], SYSERR);
		++ErrCount;
	    }
	}
    }

    /* 
     * Make sure we're not in the installation directory when we try 
     * to remove it 
     */
    if (getcwd(Dir, sizeof(Dir))) {
	chdir("/");
	Debug("Removing <%s>", Info->InstallDir);
	if (rmdir(Info->InstallDir) != 0) {
	    Error("%s: Remove installation directory failed: %s",
		  Info->InstallDir, SYSERR);
	    ++ErrCount;
	}
    } else {
	Error("Cannot determine current working directory: %s", SYSERR);
	++ErrCount;
    }

    if (ErrCount) {
	Msg(0, "\nThere were %d errors during UnInstall.\n", ErrCount);
	return -1;
    } else {
	Msg(0, "\nUnInstall succeeded.\n");
	return 0;
    }
}

/*
 * Generate name of LogFile to use
 */
static char *MakeLogFile(SwName)
     char		       *SwName;
{
    char		       *Tmp = "/var/tmp";
    char		       *cp;
    static char			File[MAXPATHLEN];

    if (File[0] || !SwName)
	return File;

#ifdef notdef /* disable for security reasons */
    if (cp = (char *) getenv("TMP"))
	Tmp = cp;
#endif

    (void) snprintf(File, sizeof(File), "%s/%s_%s_log_%d",
		    Tmp, ProgramName, SwName, getpid());

    return File;
}

/*
 * Get Distribution directory relative to McdFile
 */
extern char *GetDistDir(McdFile)
     char		       *McdFile;
{
    char		       *cp;
    char		       *Dir = NULL;

    if (!McdFile)
	return (char *) NULL;

    if (strrchr(McdFile, '/')) {
	Dir = strdup(McdFile);
	if (cp = strrchr(Dir, '/'))
	    *cp = CNULL;
    }

    return Dir;
}

/*
 * Display a line of Type
 */
static void Line(Type)
     int			Type;
{
    int				i;

    for (i = 0; i < TermWidth-1; ++i)
	Msg(0, "%c", Type);

    Msg(0, "\n");
}

/*
 * Center a line
 */
static char *Center(Msg)
     char		       *Msg;
{
    int				i;
    size_t			Len;
    static char			Buff[1024];

    Len = strlen(Msg);
    i = (TermWidth - Len) / 2;
    memset(Buff, ' ', sizeof(Buff));
    snprintf(&Buff[i], sizeof(Buff)-i, "%s", Msg);
    Buff[i + Len] = CNULL;

    return Buff;
}

/*
 * Display banner to user
 */
static void Banner(McdInfo)
     McdInfo_t		       *McdInfo;
{
    static char			Buff[512];

    Line(LT_DOUBLE);

    Msg(0, "\n");

    snprintf(Buff, sizeof(Buff), "Setup program for %s %s",
	     McdInfo->Name, McdInfo->Version);
    Msg(0, "%s\n\n", Center(Buff));

    Msg(0, "%s\n\n", Center(COPYRIGHT_MSG));

    Line(LT_DOUBLE);
}

/*
 * Ask the user to accept/decline the License
 */
static int LicenseAccept(Info, DistDir)
     McdInfo_t		       *Info;
     char		       *DistDir;
{
    if (!Info->AskForLic)
	return 0;

    if (!Info->LicFile) {
	Error("Distribution error - AskForLic is setup, but no license file was specified.");
	return -2;
    }

    Msg(0, "\nThe following is the license for this software:\n");
    Line(LT_SINGLE);
    Msg(MSGF_FILE, GetDistPath(Info->LicFile, DistDir));
    Line(LT_SINGLE);

#define LICPROMPT "Do you accept the terms of this license?"
    if (AcceptLicense) {
	Msg(0, "%s %s\n", LICPROMPT, A_ACCEPT);
	return 0;
    } else {
	Answers[0] = A_ACCEPT;
	Answers[1] = A_DECLINE;
	Answers[2] = NULL;
	Answer = Ask(LICPROMPT, Answers, NULL);
	if (!EQ(Answer, A_ACCEPT)) {
	    Msg(0, "Installation is being aborted because you declined the terms of the license.\n");
	    UserBail();
	}
    }
#undef LICPROMPT

    return 0;
}

/*
 * Install the actual files
 */
extern int InstallFiles(Mcd, Info, DistDir)
     Mcd_t		       *Mcd;
     McdInfo_t		       *Info;
     char		       *DistDir;
{
    Mcd_t		       *Entry;
    Mcd_t		       *Last = NULL;
    char		       *Src;
    static char		        Dest[MAXPATHLEN];
    struct stat			Stat;
    int				ErrCount = 0;

    memset(&MeterInfo, 0, sizeof(MeterInfo));
    MeterInfo.TermWidth = TermWidth;
    MeterInfo.Prompt = "Installing . . . ";
    for (Entry = Mcd; Entry; Entry = Entry->Next) {
	++(MeterInfo.Total);
	Last = Entry;
    }

    for (Entry = Mcd; Entry; Entry = Entry->Next) {
	++(MeterInfo.Current);
	Meter(&MeterInfo);

	snprintf(Dest, sizeof(Dest), "%s/%s", Info->InstallDir, Entry->Dest);
	if (Entry->Src)
	    Src = GetDistPath(Entry->Src, DistDir);
	else
	    Src = NULL;

	switch (Entry->Type) {
	case MCDT_FILE:
	    (void) unlink(Dest);
	    if (FLAGS_ON(Entry->Flags, MCDF_EDIT)) {
		if (CopyEdit(Src, Dest, Info) != 0)
		    Bail();
	    } else {
		if (Copy(Src, Dest) != 0)
		    Bail();
	    }
	    break;
	case MCDT_DIR:
	    if (MakeDir(Dest, Entry->Mode) != 0)
		Bail();
	    break;
	default:
	    Error("Type %d for entry \"%s\" is unknown.",
		  Entry->Type, (Entry->Src) ? Entry->Src : Entry->Dest);
	    Bail();
	}
	if (SetPerms(Dest, Entry->Owner, Entry->Group, Entry->Mode) < 0) 
	    Bail();
    }

    Msg(0, "\n");
    if (ErrCount) {
	Msg(0, "\nThere were %d errors during installation.\n", ErrCount);
	return -1;
    } else {
	return 0;
    }
}

/*
 * Return path where license file is installed
 */
extern char *GetLicensePath(Info)
     McdInfo_t		       *Info;
{
    static char			LicFile[MAXPATHLEN];

    if (Info->InstallLic[0] == '/' || Info->InstallLic[0] == '.')
	(void) snprintf(LicFile, sizeof(LicFile), "%s", Info->InstallLic);
    else
	(void) snprintf(LicFile, sizeof(LicFile), "%s/%s", 
			Info->InstallDir, Info->InstallLic);

    return LicFile;
}

/*
 * Prompt user for name of license file to be installed.
 */
extern char *LicenseFilePrompt(Info, Query)
     McdInfo_t		       *Info;
     mcl_t		       *Query;
{
    int				Status;
    int				i;

    Msg(0, "This software requires a license file for permanent operation.\n");
    Msg(0, "Without a license file the software will work properly for\n");
    Msg(0, "%d days for demonstration purposes.\n\n", DEMODAYSNUM);
    Msg(0, "You may install the license file now or later by running:\n");
    Msg(0, "    %s/%s --license --licensefile <file>\n\n",
	Info->InstallDir, BaseName(ProgramName));

    for (i = 0; i < SETUP_MAX_PROMPTS; ++i) {
	Answer = Ask("Do you wish to install the license file now?", 
		     AnswersYN, A_YES);
	if (!EQ(Answer, A_YES))
	    return (char *) NULL;
	    
	Msg(0, "The license file you name will be copied to:\n");
	Msg(0, "    %s\n", GetLicensePath(Info));
	Answer = Ask("Pathname to license file (e.g. license.mcl): (none)",
		     NULL, NULL);

	if (EQ(Answer, "none"))
	    return Answer;

	if (FileExists(Answer)) {
	    /* Valid license file */
	    Query->File = Answer;
	    Status = MCLcheck(Query);
	    if (Status != 0 || Query->Status != MCL_STAT_OK) {
		Error("%s: File does not contain a valid license.", Answer);
		continue;
	    }
	    /* Everything's fine */
	    return strdup(Answer);
	} else
	    Error("%s: The file does not exist.", Answer);
    }

    Error("Maximum Prompt limit reached.");
    return (char *) NULL;
}

/*
 * License installation
 */
extern int InstallLicense(Info, DistDir)
     McdInfo_t		       *Info;
     char		       *DistDir;
{
    static mcl_t		Query;
    int				Status;
    char		       *FileToInstall = LicenseFile;
    char		       *LicFile;

    Msg(0, "Checking license status . . . ");

    if (!Info->InstallLic) {
	Error("Distribution error - no license file location was found.");
	Bail();
    }

    LicFile = GetLicensePath(Info);

    (void) memset(&Query, 0, sizeof(Query));
    Query.File = LicFile;
    Query.Product = MCLprogramName;
    Query.ProdVers = VERSION_STR;
    if (DebugFlag)
	Query.Flags = MCLF_DEBUG;

    Status = MCLcheck(&Query);
    if (Status == 0 && (Query.Type == MCL_FULL || Query.Type == MCL_UPGRADE)) {
	Msg(0, "You already have a license.\n");
	return 0;
    }

    Msg(0, "A license file is required.\n");

    if (!FileToInstall) {
	FileToInstall = LicenseFilePrompt(Info, &Query);
	if (EQ(FileToInstall, "none"))
	    return 0;
    }

    if (!FileToInstall)
	return 0;

    Msg(0, "Installing license file %s . . . ", FileToInstall);

    if (Copy(FileToInstall, LicFile) != 0) {
	Error("Copy of %s to %s failed: %s", 
	      FileToInstall, LicFile, SYSERR);
	return -1;
    }
    if (chmod(LicFile, 0644) != 0) {
	Error("WARNING: %s: chmod failed: %s", LicFile, SYSERR);
	/* NOT FATAL */
    }

    Msg(0, "OK\n");

    return 0;
}

/*
 * Copy distirbution files into install dir
 */
extern int DistCopy(File, Info, Mode)
     char		       *File;
     McdInfo_t		       *Info;
     mode_t			Mode;
{
    static char			Path[MAXPATHLEN];

    (void) snprintf(Path, sizeof(Path), "%s/%s", 
		    Info->InstallDir, BaseName(File));

    (void) unlink(Path);

    if (Copy(File, Path) != 0) {
	Error("%s -> %s: Copy failed: %s", File, Path, SYSERR);
	return -1;
    }

    (void) SetPerms(Path, NULL, NULL, Mode);

    return 0;
}

/*
 * Installation
 */
static int Install(Mcd, Info, DistDir, MapFile, InfoFile, ProgramPath)
     Mcd_t		       *Mcd;
     McdInfo_t		       *Info;
     char		       *DistDir;
     char		       *MapFile;
     char		       *InfoFile;
     char		       *ProgramPath;
{
    static char			Path[MAXPATHLEN];

    ShowSummary(Info);
    Msg(0, "\n");
    Answer = Ask("Do you want to install this software?", AnswersYN, A_YES);
    if (!EQ(Answer, A_YES)) {
	Error("You have choosen not to proceed with installation.");
	return -1;
    }

    if (Inventory(Mcd, McdMapFile, DistDir) < 0)
	return -1;

    if (LicenseAccept(Info, DistDir) < 0)
	return -1;

    if (SysReq(Info) != 0) {
	Error("Your system does not meet the requirements for this software.");
	if (OverRide) {
	    Msg(0, "Continue? YES\n");
	} else {
	    Answer = Ask("Continue?", AnswersYN, NULL);
	    if (!ConfirmFlag && !EQ(Answer, A_YES))
		UserBail();
	}
    }

    if (InstallDir)
	Info->InstallDir = InstallDir;
    else {
	Answer = Ask("What directory do you want to install into?", 
		     NULL, Info->InstallDir);
	if (Answer)
	    Info->InstallDir = strdup(Answer);
    }

    if (CheckSpace(Info) < 0) {
	Answer = Ask("Do you wish to continue anyway?", AnswersYN, A_NO);
	if (!EQ(Answer, A_YES))
	    return -1;
    }

    Msg(0, "\n");
    ShowSummary(Info);
    Msg(0, "\n");
    Answer = Ask("Proceed and install all files now?", AnswersYN, A_YES);
    if (!EQ(Answer, A_YES)) {
	Error("You have choosen not to proceed with installation.");
	return -1;
    }

    Line(LT_SINGLE);
    Msg(0, "Preparing to install into %s . . . ", Info->InstallDir);
    MakeDir(Info->InstallDir, 0755);
    Msg(0, "OK\n");

    /*
     * Copy setup and data files to the installation dir
     */
    McdMapAddFile(Mcd, MapFile, BaseName(MapFile), 0444);
    McdMapAddFile(Mcd, InfoFile, BaseName(InfoFile), 0444);
    McdMapAddFile(Mcd, ProgramPath, BaseName(ProgramPath), 0555);
    
    if (InstallFiles(Mcd, Info, DistDir) < 0)
	return -1;

    if (InstallLicense(Info, DistDir) < 0)
	return -1;

    Msg(0, "\nInstallation is now complete.\n");

    Msg(0, "\nTo uninstall this software run the following:\n");
    Msg(0, "    %s/%s --uninstall\n\n", 
	Info->InstallDir, BaseName(ProgramPath));

    /* Show finish message */
    MsgShowFile(FN_FINISH, Info);

    return 0;
}

/*
 * The beginning
 */
main(Argc, Argv)
    int 			Argc;
    char 		      **Argv;
{    
    Mcd_t		       *McdData;
    McdInfo_t		       *Info;
    char		       *DistDir = NULL;
    char		       *cp;
    int				Status = 0;
    char		       *What = NULL;

    /*
     * Parse command line arguments.
     */
    if (ParseOptions(Opts, Num_Opts(Opts), Argc, Argv) < 0)
	exit(1);

    DistDir = GetDistDir(McdMapFile);
    if (!DistDir)
	DistDir = GetDistDir(McdInfoFile);
    if (!DistDir)
	DistDir = GetDistDir(Argv[0]);
    Debug("DistDir = <%s>", PRTS(DistDir));

    if (DebugFlag)
	Debug("%s", GetVersion());
    else if (DoPrintVersion) {
	Msg(0, "%s\n", GetVersion());
	exit(0);
    }

    Info = McdInfoRead(McdInfoFile, DistDir);
    if (!Info) {
	Error("Cannot find MCDINFO.");
	exit(1);
    }

    if (cp = MakeLogFile(Info->Name)) {
	SetLogFile(cp);
	Msg(0, "Logging results to %s\n", cp);
    }

    Banner(Info);

    McdData = McdMapRead(McdMapFile, DistDir);
    if (!McdData)
	exit(1);

    if (!OpInstall && !OpLicense && !OpUnInstall && !OpVerify && !OpInventory)
	OpInstall++;

    if (!OpInstall && DistDir)
	Info->InstallDir = DistDir;

    What = "Undefined operation";
    Status = -1;

    if (OpInstall) {
	Status = Install(McdData, Info, DistDir, 
			 McdMapFile, McdInfoFile, Argv[0]);
	What = "Installation";
    } else if (OpUnInstall) {
	Status = UnInstall(McdData, Info, DistDir, Argv[0]);
	What = "UnInstall";
    } else if (OpInventory) {
	Status = Inventory(McdData, McdMapFile, DistDir);
	What = "Inventory";
    } else if (OpVerify) {
	Status = Verify(McdData, Info, DistDir, Argv[0]);
	What = "Verify";
    } else if (OpLicense) {
	Status = InstallLicense(Info, DistDir);
	What = "License installation";
    }

    if (Status < 0)
	Error("%s failed.", What);
    Msg(0, "A log of this %s action is available as\n    %s\n", 
	What, MakeLogFile(NULL));

    exit(Status);
}
